"""Routines for managing SQL statements"""
import logging
import re
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

from psycopg2.sql import SQL, Composed, Identifier

from splitgraph.config import SPLITGRAPH_META_SCHEMA
from splitgraph.core.sql._validation import (
    IMPORT_SQL_PERMITTED_STATEMENTS,
    PG_CATALOG_TABLES,
    SPLITFILE_SQL_PERMITTED_STATEMENTS,
)

try:
    import pglast.node
    from pglast import parse_sql
    from pglast.node import Node, Scalar
    from pglast.parser import ParseError
    from pglast.stream import IndentedStream

    _VALIDATION_SUPPORTED = True
except ImportError:
    _VALIDATION_SUPPORTED = False

from splitgraph.exceptions import UnsupportedSQLError

# https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
POSTGRES_MAX_IDENTIFIER = 63


def _validate_range_var(node: "Node") -> None:
    if node.ast_node.schemaname:
        raise UnsupportedSQLError("Table names must not be schema-qualified!")
    if node.ast_node.relname in PG_CATALOG_TABLES:
        raise UnsupportedSQLError("Invalid table name %s!" % node.ast_node.relname)


def _validate_funccall(node: "Node"):
    # We can't ban all function calls (there are some useful ones like
    # to_date() etc) but we can't allow all of them either (some pg system
    # calls like pg_relation_filepath etc) aren't appropriate here but are available
    # to all users by default -- so as a preliminary defense we stop all functions that
    # begin with pg_.
    funcname = node.funcname
    if len(funcname) != 1:
        # e.g. pg_catalog.substring
        funcname = funcname[1]
    if isinstance(funcname, pglast.node.List):
        funcname_s = funcname.string_value
    elif funcname.node_tag == "String":
        funcname_s = funcname.val.value
    else:
        raise AssertionError
    if funcname_s.startswith("pg_"):
        raise UnsupportedSQLError("Unsupported function name %s!" % funcname)


# Nodes in this list have extra validators that are supposed to return None or raise an Exception if they
# fail validation.
_SQL_VALIDATORS = {"FuncCall": _validate_funccall}
_IMPORT_SQL_VALIDATORS = {"RangeVar": _validate_range_var, "FuncCall": _validate_funccall}

# Fallback regex to do best-effort SQL rewriting on Windows where pglast doesn't work.
_SCHEMA_RE = re.compile(r'"([\w-]+/)?([\w-]+):([\w-]+)"')


def _validate_node(
    node: Union["Scalar", "Node"],
    permitted_statements: List[str],
    node_validators: Dict[str, Callable],
) -> None:
    if isinstance(node, Scalar):
        return
    node_class = node.node_tag
    if node_class in node_validators:
        node_validators[node_class](node)
    elif node_class.endswith("Stmt") and node_class not in permitted_statements:
        message = "Unsupported statement type %s" % node_class
        if "location" in node.attribute_names and node.ast_node.location:
            message += " near character %d" % node.ast_node.location
        raise UnsupportedSQLError(message + "!")


def _emit_ast(ast: "Node") -> str:
    # required to instantiate all pglast node printers.
    # noinspection PyUnresolvedReferences
    from pglast import printers  # noqa

    stream = IndentedStream()
    return str(stream(ast))


def recover_original_schema_name(sql: str, schema_name: str) -> str:
    """Postgres truncates identifiers to 63 characters at parse time and, as pglast
    uses bits of PG to parse queries, image names like noaa/climate:64_chars_of_hash
    get truncated which can cause ambiguities and issues in provenance. We can't
    get pglast to give us back the full identifier, but we can try and figure out
    what it used to be and patch the AST to have it again.
    """
    if len(schema_name) < POSTGRES_MAX_IDENTIFIER:
        return schema_name

    candidates = list(set(re.findall(r"(" + re.escape(schema_name) + r"[^.\"]*)[.\"]", sql)))
    # Us finding more than one candidate schema is pretty unlikely to happen:
    # we'd have to have a truncated schema name that's 63 characters long
    # (of kind some_namespace/some_repo:abcdef1234567890....)
    # which also somehow features in this query as a non-identifier. Raise an error here if
    # this does happen.
    assert len(candidates) == 1
    return str(candidates[0])


def _rewrite_sql_fallback(sql: str, image_mapper: Callable) -> Tuple[str, str]:
    from splitgraph.core.repository import Repository

    replacements: List[Tuple[str, str, str]] = []

    for namespace, repository, image_hash in _SCHEMA_RE.findall(sql):
        original_text = namespace + repository + ":" + image_hash
        if namespace:
            # Strip the / at the end
            namespace = namespace[:-1]

        temporary_schema, canonical_name = image_mapper(
            Repository(namespace, repository), image_hash
        )
        replacements.append((original_text, temporary_schema, canonical_name))

    canonical_sql = sql
    rewritten_sql = sql

    for original_text, temporary_schema, canonical_name in replacements:
        rewritten_sql = rewritten_sql.replace(original_text, temporary_schema)
        canonical_sql = canonical_sql.replace(original_text, canonical_name)

    return rewritten_sql, canonical_sql


def prepare_splitfile_sql(sql: str, image_mapper: Callable) -> Tuple[str, str]:
    """
    Transform an SQL query to prepare for it to be used in a Splitfile SQL command and validate it.
    The rules are:

      * Only basic DDL (CREATE/ALTER/DROP table) and DML (SELECT/INSERT/UPDATE/DELETE) are permitted.
      * All tables must be either non-schema qualified (the statement is run with `search_path`
      set to the single schema that a Splitgraph image is checked out into) or have schemata of
      format namespace/repository:hash_or_tag. In the second case, the schema is rewritten to point
      at a temporary mount of the Splitgraph image.

    :param sql: SQL query
    :param image_mapper: Takes in an image and gives back the schema it should be rewritten to
        (for the purposes of execution) and the canonical form of the image.
    :return: Transformed form of the SQL with substituted schema shims for Splitfile execution
        and the canonical form (with e.g. tags resolved into at-the-time full image hashes)
    :raises: UnsupportedSQLException if validation failed
    """

    if not _VALIDATION_SUPPORTED:
        logging.warning("SQL validation is unsupported on Windows. SQL will be run unvalidated.")

        return _rewrite_sql_fallback(sql, image_mapper)

    # Avoid circular import
    from splitgraph.core.output import parse_repo_tag_or_hash

    try:
        tree = Node(parse_sql(sql))
    except ParseError as e:
        raise UnsupportedSQLError("Could not parse %s: %s" % (sql, str(e)))

    # List of dict pointers (into parts of the AST) and new schema names we have
    # to rewrite them to. We need to emit two kinds of rewritten SQL: one with schemata
    # replaced with LQ shims (that we send to PostgreSQL for execution) and one with
    # schemata replaced with full image names that we store in provenance_data for reruns etc.
    # On the first pass, we rewrite the parse tree to have the first kind of schemata, then
    # get pglast to serialize it, then use this dictionary to rewrite just the interesting
    # parts of the parse tree (instead of having to re-traverse/re-crawl the tree).
    future_rewrites = []

    for node in tree.traverse():
        _validate_node(
            node,
            permitted_statements=SPLITFILE_SQL_PERMITTED_STATEMENTS,
            node_validators=_SQL_VALIDATORS,
        )

        if not isinstance(node, Node) or node.node_tag != "RangeVar":
            continue

        ast_node = node.ast_node

        if ast_node.relname in PG_CATALOG_TABLES:
            raise UnsupportedSQLError("Invalid table name %s!" % ast_node.relname)
        if not ast_node.schemaname:
            continue

        schema_name = recover_original_schema_name(sql, ast_node.schemaname)

        # If the table name is schema-qualified, rewrite it to talk to a LQ shim
        repo, hash_or_tag = parse_repo_tag_or_hash(schema_name, default="latest")
        temporary_schema, canonical_name = image_mapper(repo, hash_or_tag)

        # We have to access the internal parse tree here to rewrite the schema.
        ast_node.schemaname = temporary_schema
        future_rewrites.append((ast_node, canonical_name))

    rewritten_sql = _emit_ast(tree)

    for tree_subset, canonical_name in future_rewrites:
        tree_subset.schemaname = canonical_name

    canonical_sql = _emit_ast(tree)
    return rewritten_sql, canonical_sql


def validate_import_sql(sql: str) -> str:
    """
    Check an SQL query to see if it can be safely used in an IMPORT statement
    (e.g. `FROM noaa/climate:latest IMPORT {SELECT * FROM rainfall WHERE state = 'AZ'} AS rainfall`.
    In this case, only a single SELECT statement is supported.

    :param sql: SQL query
    :return: Canonical (formatted) form of the SQL statement
    :raises: UnsupportedSQLException if validation failed
    """
    if not _VALIDATION_SUPPORTED:
        logging.warning("SQL validation is unsupported on Windows. SQL will be run unvalidated.")
        return sql

    try:
        tree = Node(parse_sql(sql))
    except ParseError as e:
        raise UnsupportedSQLError("Could not parse %s: %s" % (sql, str(e)))
    if len(tree) != 1:
        raise UnsupportedSQLError("The query is supposed to consist of only one SELECT statement!")

    for node in tree.traverse():
        _validate_node(
            node,
            permitted_statements=IMPORT_SQL_PERMITTED_STATEMENTS,
            node_validators=_IMPORT_SQL_VALIDATORS,
        )
    return _emit_ast(tree)


def select(
    table: str,
    columns: str = "*",
    where: str = "",
    schema: str = SPLITGRAPH_META_SCHEMA,
    table_args: Optional[str] = None,
) -> Composed:
    """
    A generic SQL SELECT constructor to simplify metadata access queries so that we don't have to repeat the same
    identifiers everywhere.

    :param table: Table to select from.
    :param columns: Columns to select as a string. WARN: concatenated directly without any formatting.
    :param where: If specified, added to the query with a "WHERE" keyword. WARN also concatenated directly.
    :param schema: Defaults to SPLITGRAPH_META_SCHEMA.
    :param table_args: If specified, appends to the FROM clause after the table specification,
        for example, SELECT * FROM "splitgraph_api"."get_images" (%s, %s) ...
    :return: A psycopg2.sql.SQL object with the query.
    """
    query = SQL("SELECT " + columns) + SQL(" FROM {}.{}").format(
        Identifier(schema), Identifier(table)
    )
    if table_args:
        query += SQL(table_args)
    if where:
        query += SQL(" WHERE " + where)
    return query


def insert(table: str, columns: Sequence[str], schema: str = SPLITGRAPH_META_SCHEMA) -> Composed:
    """
    A generic SQL SELECT constructor to simplify metadata access queries so that we don't have to repeat the same
    identifiers everywhere.

    :param table: Table to select from.
    :param columns: Columns to insert as a list of strings.
    :param schema: Schema that contains the table
    :return: A psycopg2.sql.SQL object with the query (parameterized)
    """
    query = SQL("INSERT INTO {}.{}").format(Identifier(schema), Identifier(table))
    query += SQL("(" + ",".join("{}" for _ in columns) + ")").format(*map(Identifier, columns))
    query += SQL("VALUES (" + ",".join("%s" for _ in columns) + ")")
    return query



from abc import ABC, abstractmethod
from contextlib import contextmanager
from importlib.resources import read_binary
from random import getrandbits
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, cast

from psycopg2._json import Json
from psycopg2.sql import SQL, Identifier

from splitgraph.config import DEFAULT_CHUNK_SIZE
from splitgraph.core.engine import repository_exists
from splitgraph.core.image import Image
from splitgraph.core.image_mounting import DefaultImageMounter, ImageMounter
from splitgraph.core.repository import Repository
from splitgraph.core.types import (
    Credentials,
    IntrospectionResult,
    MountError,
    Params,
    SyncState,
    TableColumn,
    TableInfo,
)
from splitgraph.engine import ResultShape
from splitgraph.ingestion.common import add_timestamp_tags
from splitgraph.resources import icons

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine

INGESTION_STATE_TABLE = "_sg_ingestion_state"
INGESTION_STATE_SCHEMA = [
    TableColumn(1, "timestamp", "timestamp", True, None),
    TableColumn(2, "state", "json", False, None),
]


class DataSource(ABC):
    params_schema: Dict[str, Any] = {"type": "object"}
    credentials_schema: Dict[str, Any] = {"type": "object"}
    table_params_schema: Dict[str, Any] = {"type": "object"}

    supports_mount = False
    supports_sync = False
    supports_load = False
    _icon_file: Optional[str] = None

    @classmethod
    def get_icon(cls) -> Optional[bytes]:
        if cls._icon_file:
            return read_binary(icons, cls._icon_file)
        return None

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_description(cls) -> str:
        raise NotImplementedError

    def __init__(
        self,
        engine: "PostgresEngine",
        credentials: Credentials,
        params: Params,
        tables: Optional[TableInfo] = None,
    ):
        import jsonschema

        self.engine = engine

        if "tables" in params:
            tables = params.pop("tables")

        jsonschema.validate(instance=credentials, schema=self.credentials_schema)
        jsonschema.validate(instance=params, schema=self.params_schema)

        self.credentials = credentials
        self.params = params

        self._validate_table_params(tables)
        self.tables = tables

    @classmethod
    def _validate_table_params(cls, tables: Optional[TableInfo]) -> None:
        import jsonschema

        if isinstance(tables, dict):
            for _, table_params in tables.values():
                jsonschema.validate(instance=table_params, schema=cls.table_params_schema)

    @abstractmethod
    def introspect(self) -> IntrospectionResult:
        raise NotImplementedError

    def get_raw_url(
        self, tables: Optional[TableInfo] = None, expiry: int = 3600
    ) -> Dict[str, List[Tuple[str, str]]]:
        """
        Get a list of public URLs for each table in this data source, e.g. to export the data
        as CSV. These may be temporary (e.g. pre-signed S3 URLs) but should be accessible without
        authentication.
        :param tables: A TableInfo object overriding the table params of the source
        :param expiry: The URL should be valid for at least this many seconds
        :return: Dict of table_name -> list of (mimetype, raw URL)
        """
        return {}


class MountableDataSource(DataSource, ABC):
    supports_mount = True

    @abstractmethod
    def mount(
        self,
        schema: str,
        tables: Optional[TableInfo] = None,
        overwrite: bool = True,
    ) -> Optional[List[MountError]]:
        """Instantiate the data source as foreign tables in a schema"""
        raise NotImplementedError


class LoadableDataSource(DataSource, ABC):
    supports_load = True

    @abstractmethod
    def _load(self, schema: str, tables: Optional[TableInfo] = None):
        raise NotImplementedError

    def load(self, repository: "Repository", tables: Optional[TableInfo] = None) -> str:
        self._validate_table_params(tables)
        if not repository_exists(repository):
            repository.init()

        image_hash = "{:064x}".format(getrandbits(256))
        tmp_schema = "{:064x}".format(getrandbits(256))
        repository.images.add(
            parent_id=None,
            image=image_hash,
        )
        repository.object_engine.create_schema(tmp_schema)

        try:
            self._load(schema=tmp_schema, tables=tables)

            repository._commit(
                head=None,
                image_hash=image_hash,
                snap_only=True,
                chunk_size=DEFAULT_CHUNK_SIZE,
                schema=tmp_schema,
            )
            add_timestamp_tags(repository, image_hash)
        except BaseException:
            repository.rollback_engines()
            repository.images.delete([image_hash])
            repository.commit_engines()
            raise
        finally:
            repository.object_engine.delete_schema(tmp_schema)
            repository.commit_engines()

        return image_hash


class SyncableDataSource(LoadableDataSource, DataSource, ABC):
    supports_load = True
    supports_sync = True

    @abstractmethod
    def _sync(
        self, schema: str, state: Optional[SyncState] = None, tables: Optional[TableInfo] = None
    ) -> Optional[SyncState]:
        """Incremental load"""
        raise NotImplementedError

    def sync(
        self,
        repository: "Repository",
        image_hash: Optional[str],
        tables: Optional[TableInfo] = None,
    ) -> str:
        self._validate_table_params(tables)
        if not repository_exists(repository):
            repository.init()

        state = get_ingestion_state(repository, image_hash)
        image_hash = image_hash or "0" * 64
        repository.images[image_hash].checkout()

        try:
            new_state = self._sync(schema=repository.to_schema(), state=state, tables=tables)

            if new_state:
                # Write the new state to the table
                if not repository.object_engine.table_exists(
                    repository.to_schema(), INGESTION_STATE_TABLE
                ):
                    repository.object_engine.create_table(
                        repository.to_schema(), INGESTION_STATE_TABLE, INGESTION_STATE_SCHEMA
                    )

                repository.run_sql(
                    SQL("INSERT INTO {} (timestamp, state) VALUES(now(), %s)").format(
                        Identifier(INGESTION_STATE_TABLE)
                    ),
                    (Json(new_state),),
                )

            new_image = repository.commit()
        finally:
            repository.uncheckout()
            repository.commit_engines()

        return new_image.image_hash

    def _load(self, schema: str, tables: Optional[TableInfo] = None):
        self._sync(schema, tables=tables, state=None)


class TransformingDataSource(DataSource, ABC):
    """
    Data source that runs transformations between Splitgraph images. Takes in an extra
    parameter, an ImageMounter instance to manage temporary image checkouts.
    """

    def __init__(
        self,
        engine: "PostgresEngine",
        credentials: Credentials,
        params: Params,
        tables: Optional[TableInfo] = None,
        image_mounter: Optional[ImageMounter] = None,
    ):
        super().__init__(engine, credentials, params, tables)
        self._mounter = image_mounter or DefaultImageMounter(engine)

    @abstractmethod
    def get_required_images(self) -> List[Tuple[str, str, str]]:
        """
        Get images required by this data source.
        :returns List of tuples (namespace, repository, hash_or_tag)
        """
        raise NotImplementedError

    @contextmanager
    def mount_required_images(self) -> Generator[Dict[Tuple[str, str, str], str], None, None]:
        """
        Mount all images required by this data source into temporary schemas. On exit from this
        context manager, unmounts them.
        :return: Map of (namespace, repository, hash_or_tag) -> schema where the image is mounted.
        """
        try:
            schema_map = self._mounter.mount(self.get_required_images())
            yield schema_map
        finally:
            self._mounter.unmount()


def get_ingestion_state(repository: "Repository", image_hash: Optional[str]) -> Optional[SyncState]:
    state = None

    if image_hash:
        image = repository.images[image_hash]
        if INGESTION_STATE_TABLE in image.get_tables():
            with image.query_schema() as s:
                state = repository.object_engine.run_sql(
                    SQL("SELECT state FROM {}.{} ORDER BY timestamp DESC LIMIT 1").format(
                        Identifier(s), Identifier(INGESTION_STATE_TABLE)
                    ),
                    return_shape=ResultShape.ONE_ONE,
                )
    return cast(Optional[SyncState], state)


def prepare_new_image(
    repository: "Repository", hash_or_tag: Optional[str], comment: str = "Singer tap ingestion"
) -> Tuple[Optional[Image], str]:
    new_image_hash = "{:064x}".format(getrandbits(256))
    if repository_exists(repository):
        # Clone the base image and delta compress against it
        base_image: Optional[Image] = repository.images[hash_or_tag] if hash_or_tag else None
        repository.images.add(parent_id=None, image=new_image_hash, comment=comment)
        if base_image:
            repository.engine.run_sql(
                "INSERT INTO splitgraph_meta.tables "
                "(SELECT namespace, repository, %s, table_name, table_schema, object_ids "
                "FROM splitgraph_meta.tables "
                "WHERE namespace = %s AND repository = %s AND image_hash = %s)",
                (
                    new_image_hash,
                    repository.namespace,
                    repository.repository,
                    base_image.image_hash,
                ),
            )
    else:
        base_image = None
        repository.images.add(parent_id=None, image=new_image_hash, comment=comment)
    return base_image, new_image_hash



"""Default Splitgraph engine: uses PostgreSQL to store metadata and actual objects and an audit stored procedure
to track changes, as well as the Postgres FDW interface to upload/download objects to/from other Postgres engines."""
import importlib.resources
import itertools
import json
import logging
import sys
import time
from contextlib import contextmanager
from io import BytesIO, TextIOWrapper
from pathlib import PurePosixPath
from threading import get_ident
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import psycopg2
import psycopg2.extensions
from packaging.version import Version
from psycopg2 import DatabaseError
from psycopg2.errors import InvalidSchemaName, UndefinedTable
from psycopg2.extras import Json
from psycopg2.pool import AbstractConnectionPool, ThreadedConnectionPool
from psycopg2.sql import SQL, Composed, Identifier
from tqdm import tqdm

from splitgraph.__version__ import __version__
from splitgraph.config import (
    CONFIG,
    SG_CMD_ASCII,
    SPLITGRAPH_API_SCHEMA,
    SPLITGRAPH_META_SCHEMA,
)
from splitgraph.core import server
from splitgraph.core.common import META_TABLES, ensure_metadata_schema
from splitgraph.core.sql import select
from splitgraph.core.types import TableColumn, TableSchema
from splitgraph.engine import (
    ChangeEngine,
    ObjectEngine,
    ResultShape,
    SQLEngine,
    switch_engine,
    validate_type,
)
from splitgraph.exceptions import (
    APICompatibilityError,
    AuthAPIError,
    EngineInitializationError,
    IncompleteObjectDownloadError,
    ObjectMountingError,
    ObjectNotFoundError,
)
from splitgraph.hooks.mount_handlers import mount_postgres
from splitgraph.resources import static

if TYPE_CHECKING:
    # Import the connection object under a different name as it shadows
    # the connection property otherwise
    from psycopg2._psycopg import connection as Connection

psycopg2.extensions.register_adapter(dict, Json)

_AUDIT_SCHEMA = "splitgraph_audit"
_AUDIT_TRIGGER = "audit_trigger.sql"
_PUSH_PULL = "splitgraph_api.sql"
_CSTORE = "cstore.sql"
CSTORE_SERVER = "cstore_server"
_PACKAGE = "splitgraph"
ROW_TRIGGER_NAME = "audit_trigger_row"
STM_TRIGGER_NAME = "audit_trigger_stm"
REMOTE_TMP_SCHEMA = "tmp_remote_data"
SG_UD_FLAG = "sg_ud_flag"

# Retry policy for connection errors
RETRY_DELAY = 5
RETRY_AMOUNT = 12

# Internal API data
_API_VERSION = "0.0.1"

# Limitations for SQL API that the client uses to talk to the registry. Because
# we let the client run SQL in a controlled environment on the registry, it allows
# us to batch query calls or add custom filters without adding a new API call,
# however we'd still like to limit what the client can do to make each API's
# footprint more predictable. We hence chunk some heavy queries like object
# index uploading to allow registering multiple objects at the same time whilst
# also limiting the query's maximum size.
API_MAX_QUERY_LENGTH = 261000

# In addition, some API calls (like get_object_meta) allow variadic arguments
# to decrease the number of roundtrips the client has to do -- limit this to a sane
# number.
API_MAX_VARIADIC_ARGS = 100


# PG types we can run max/min/comparisons on

PG_INDEXABLE_TYPES = [
    "bigint",
    "bigserial",
    "bit",
    "character",
    "character varying",
    "cidr",
    "date",
    "double precision",
    "inet",
    "integer",
    "money",
    "numeric",
    "real",
    "smallint",
    "smallserial",
    "serial",
    "text",
    "time",
    "time without time zone",
    "time with time zone",
    "timestamp",
    "timestamp without time zone",
    "timestamp with time zone",
]


def _quiet():
    # Don't spam connection error messages if we're in interactive mode and
    # loglevel is default (INFO/WARNING).
    return sys.stdin.isatty() and logging.getLogger().getEffectiveLevel() >= logging.INFO


def _handle_fatal(e):
    """Handle some Postgres exceptions that aren't transient."""

    if "unexpected response from login query" in str(e):
        # This one is unintuitive but is raised by the sg registry auth gateway
        # if there's been a login error (user doesn't exist or wrong password)
        raise AuthAPIError(
            "Could not open a registry connection: wrong API key or secret. "
            "Check your credentials using sgr config and make sure you've "
            "logged into the registry using sgr cloud login."
        )
    if "password authentication failed" in str(e):
        raise e


def _paginate_by_size(cur, query, argslist, max_size=API_MAX_QUERY_LENGTH):
    """Take a query and a list of args and return a list of binary strings
    (each shorter than max_size) chunking the query up into batches.
    """
    buf = b""
    for args in argslist:
        statement = cur.mogrify(query, args)
        # If max_size is <= 0, don't split the argslist up
        if 0 < max_size < len(statement):
            raise ValueError(
                ("Statement %s... exceeds maximum query size %d. " % (statement[:100], max_size))
                + "Are you trying to upload an object with a large amount of metadata?"
            )
        if 0 < max_size < len(statement) + len(buf) + 1:
            yield buf
            buf = b""
        buf += statement + b";"

    if buf:
        yield buf


T = TypeVar("T")


def chunk(sequence: Sequence[T], chunk_size: int = API_MAX_VARIADIC_ARGS) -> Iterator[List[T]]:
    curr_chunk: List[T] = []
    for i, curr in enumerate(sequence):
        curr_chunk.append(curr)
        if (i + 1) % chunk_size == 0:
            yield curr_chunk
            curr_chunk = []
    if curr_chunk:
        yield curr_chunk


def _get(d: Dict[str, Optional[str]], k: str) -> str:
    result = d.get(k)
    assert result
    return result


def get_conn_str(conn_params: Dict[str, Optional[str]]) -> str:
    server, port, username, password, dbname = (
        _get(conn_params, "SG_ENGINE_HOST"),
        _get(conn_params, "SG_ENGINE_PORT"),
        _get(conn_params, "SG_ENGINE_USER"),
        _get(conn_params, "SG_ENGINE_PWD"),
        _get(conn_params, "SG_ENGINE_DB_NAME"),
    )
    return f"postgresql://{username}:{password}@{server}:{port}/{dbname}"


def _quote_ident(val: str) -> str:
    return '"%s"' % val.replace('"', '""')


class PsycopgEngine(SQLEngine):
    """Postgres SQL engine backed by a Psycopg connection."""

    def __init__(
        self,
        name: Optional[str],
        conn_params: Optional[Dict[str, Optional[str]]] = None,
        pool: Optional[AbstractConnectionPool] = None,
        autocommit: bool = False,
        registry: bool = False,
        in_fdw: bool = False,
        check_version: bool = True,
    ) -> None:
        """
        :param name: Name of the engine
        :param conn_params: Optional, dictionary of connection params as stored in the config.
        :param pool: If specified, a Psycopg connection pool to use in this engine. By default, parameters
            in conn_params are used so one of them must be specified.
        :param autocommit: If True, the engine will not use transactions for its operation.
        """
        super().__init__()

        if not conn_params and not pool:
            raise ValueError("One of conn_params/pool must be specified!")

        self.name = name
        self.autocommit = autocommit
        self.connected = False
        self.check_version = check_version
        self.registry = registry
        self.in_fdw = in_fdw

        """List of notices issued by the server during the previous execution of run_sql."""
        self.notices: List[str] = []

        if conn_params:
            self.conn_params = conn_params

            # Connection pool used by the engine, keyed by the thread ID (so one connection gets
            # claimed per thread). Usually, only one connection is used (for e.g. metadata management
            # or performing checkouts/commits). Multiple connections are used when downloading/uploading
            # objects from S3, since that is done in multiple threads and these connections are short-lived.
            server, port, username, password, dbname = (
                conn_params["SG_ENGINE_HOST"],
                conn_params["SG_ENGINE_PORT"],
                conn_params["SG_ENGINE_USER"],
                conn_params["SG_ENGINE_PWD"],
                conn_params["SG_ENGINE_DB_NAME"],
            )

            self._pool = ThreadedConnectionPool(
                minconn=0,
                maxconn=conn_params.get("SG_ENGINE_POOL", CONFIG["SG_ENGINE_POOL"]),
                host=server,
                port=int(cast(int, port)) if port is not None else None,
                user=username,
                password=password,
                dbname=dbname,
                application_name="sgr " + __version__,
            )
        else:
            self._pool = pool

    def __repr__(self) -> str:
        try:
            conn_summary = "(%s@%s:%s/%s)" % (
                self.conn_params["SG_ENGINE_USER"],
                self.conn_params["SG_ENGINE_HOST"],
                self.conn_params["SG_ENGINE_PORT"],
                self.conn_params["SG_ENGINE_DB_NAME"],
            )
        except AttributeError:
            conn_summary = " (custom connection pool)"

        return "PostgresEngine " + ((self.name + " ") if self.name else "") + conn_summary

    def commit(self) -> None:
        if self.connected:
            conn = self.connection
            conn.commit()
            self._pool.putconn(conn)

    def close_others(self) -> None:
        """
        Close and release all other connections to the connection pool.
        """
        # This is here because of the way ThreadPoolExecutor works: from within the function
        # that it calls, we have no way of knowing if we'll be called again in that thread.
        # We also can't release the connection back into the pool because that causes the
        # pool to reset it, clearing all state that we might need. Hence this has to be
        # called after the TPE has finished.

        if self.connected:
            us = get_ident()
            other_conns = [v for k, v in self._pool._used.items() if k != us]
            for c in other_conns:
                self._pool.putconn(c, close=True)

    def close(self) -> None:
        if self.connected:
            conn = self.connection
            conn.close()
            self._pool.putconn(conn)

    def rollback(self) -> None:
        if self.connected:
            if self._savepoint_stack.stack:
                self.run_sql(SQL("ROLLBACK TO ") + Identifier(self._savepoint_stack.stack.pop()))
            else:
                conn = self.connection
                conn.rollback()
                self._pool.putconn(conn)

    def lock_table(self, schema: str, table: str) -> None:
        # Allow SELECTs but not writes to a given table.
        self.run_sql(
            SQL("LOCK TABLE {}.{} IN EXCLUSIVE MODE").format(Identifier(schema), Identifier(table))
        )

    @contextmanager
    def copy_cursor(self):
        """
        Return a cursor that can be used for copy_expert operations
        """

        # Suspend the callback that terminates the connection on Ctrl+C since it doesn't work
        # with copy_expert: https://stackoverflow.com/a/37606536
        thread = psycopg2.extensions.get_wait_callback()
        try:
            psycopg2.extensions.set_wait_callback(None)
            with self.connection.cursor() as cur:
                yield cur
        finally:
            psycopg2.extensions.set_wait_callback(thread)

    @property
    def splitgraph_version(self) -> Optional[str]:
        """Returns the version of the Splitgraph library installed on the engine
        and by association the version of the engine itself."""
        if self.registry:
            raise ValueError("Engine %s is a registry, can't check version!" % self.name)
        else:
            return self._call_version_func(self.connection)

    @property
    def connection(self) -> "Connection":
        """Engine-internal Psycopg connection."""
        retries = 0
        failed = False

        def _notify():
            nonlocal failed
            print(("\033[F" if failed else "") + "Waiting for connection..." + "." * retries)
            sys.stdout.flush()
            failed = True

        while True:
            try:
                conn = self._pool.getconn(get_ident())
                if conn.closed:
                    self._pool.putconn(conn)
                    conn = self._pool.getconn(get_ident())
                # Flip the autocommit flag if needed: this is a property,
                # so changing it from False to False will fail if we're actually in a transaction.
                if conn.autocommit != self.autocommit:
                    conn.autocommit = self.autocommit

                if not self.connected:
                    # Get the engine version and warn the user if the client and the
                    # engine have mismatched versions.
                    if not self.registry and self.check_version:
                        self._check_engine_version(conn)

                    # If we're connecting to a registry, check the API
                    # version and warn about incompatibilities.
                    if self.registry:
                        self._check_api_compat(conn)

                self.connected = True
                return conn
            # If we can't connect right now, write/log a message and wait. This is slightly
            # hacky since we don't get passed whether we've been called through Click, so
            # we check if we're running interactively and then print some
            # dots instead of a scary log message (waiting until the connection actually fails)
            except psycopg2.errors.OperationalError as e:
                if retries >= RETRY_AMOUNT:
                    logging.exception(
                        "Error connecting to the engine after %d retries", RETRY_AMOUNT
                    )
                    raise
                _handle_fatal(e)
                retries += 1
                if _quiet():
                    _notify()
                else:
                    logging.error(
                        "Error connecting to the engine (%s: %s), "
                        "sleeping %.2fs and retrying (%d/%d)...",
                        str(e.__class__.__name__),
                        e,
                        RETRY_DELAY,
                        retries,
                        RETRY_AMOUNT,
                    )
                time.sleep(RETRY_DELAY)

    def _call_version_func(self, conn, func="get_splitgraph_version") -> Optional[str]:
        # Internal function to get the Splitgraph library/API version on the engine.
        # Doesn't use run_sql because it can get called as part of run_sql if it's
        # run for the first time, messing up transaction state.
        try:
            with conn.cursor() as cur:
                cur.execute(
                    SQL("SELECT {}.{}()").format(
                        Identifier(SPLITGRAPH_API_SCHEMA),
                        Identifier(func),
                    )
                )
                result = cur.fetchone()
                if result:
                    return cast(str, result[0])
                return None
        except psycopg2.errors.UndefinedFunction:
            conn.rollback()
            return None
        except psycopg2.errors.InvalidSchemaName:
            conn.rollback()
            return None

    def _check_engine_version(self, conn):
        engine_version = self._call_version_func(conn)
        if engine_version and engine_version != __version__:
            logging.warning(
                "You're running sgr %s against engine %s. "
                "It's recommended to match the versions of sgr and the "
                "engine. Either upgrade your engine or ignore this message "
                "by setting SG_CHECK_VERSION= to this engine's config.",
                __version__,
                engine_version,
            )

    def _check_api_compat(self, conn):
        # TODO here: store the version
        #   in imagemanager: check the api version, call different get_tag (more args)
        #
        remote_version = self._call_version_func(conn, "get_version")
        if not remote_version:
            return

        client = Version(_API_VERSION)
        remote = Version(remote_version)

        if client.major != remote.major:
            raise APICompatibilityError(
                "Incompatible API version between client (%s) and registry %s (%s)!"
                % (_API_VERSION, self.name, remote_version)
            )

        if client.minor != remote.minor:
            logging.warning(
                "Client has a different API version (%s) than the registry %s (%s)."
                % (_API_VERSION, self.name, remote_version)
            )

    def run_chunked_sql(
        self,
        statement: Union[bytes, Composed, str, SQL],
        arguments: Sequence[Any],
        return_shape: Optional[ResultShape] = ResultShape.MANY_MANY,
        chunk_size: int = API_MAX_VARIADIC_ARGS,
        chunk_position: int = -1,
    ) -> Any:
        """Because the Splitgraph API has a request size limitation, certain
        SQL calls with variadic arguments are going to be too long to fit that. This function
        runs an SQL query against a set of broken up arguments and returns the combined result.
        """

        # Calculate arguments
        batches: Iterable[Any]
        if chunk_position == -1:
            # The whole arguments sequence is chunked up (treated as a list of tuples)
            batches = chunk(arguments, chunk_size)
        else:
            # Treat the nth argument as a list and chunk in that instead
            subbatches = chunk(arguments[chunk_position], chunk_size)
            batches = [
                tuple(arguments[:chunk_position]) + (s,) + tuple(arguments[chunk_position + 1 :])
                for s in subbatches
            ]

        results = [self.run_sql(statement, batch, return_shape) for batch in batches]

        # Join up the results -- we can only have one row-many cols (a list of lists of singletons)
        # or many rows-many cols (a list of lists of tuples) here
        return [r for rs in results if rs for r in rs]

    def run_sql(
        self,
        statement: Union[bytes, Composed, str, SQL],
        arguments: Optional[Sequence[Any]] = None,
        return_shape: Optional[ResultShape] = ResultShape.MANY_MANY,
        named: bool = False,
    ) -> Any:

        cursor_kwargs = {"cursor_factory": psycopg2.extras.NamedTupleCursor} if named else {}
        connection = self.connection

        attempt = 0

        while True:
            with connection.cursor(**cursor_kwargs) as cur:
                try:
                    self.notices = []
                    cur.execute(statement, arguments)
                    if connection.notices:
                        self.notices = connection.notices[:]
                        del connection.notices[:]

                        if self.registry:
                            # Forward NOTICE messages from the registry back to the user
                            # (e.g. to nag them to upgrade etc).
                            for notice in self.notices:
                                logging.info("%s says: %s", self.name, notice)
                except Exception as e:
                    # Rollback the transaction (to a savepoint if we're inside the savepoint() context manager)
                    self.rollback()
                    # Go through some more common errors (like the engine not being initialized) and raise
                    # more specific Splitgraph exceptions.
                    if isinstance(e, UndefinedTable):
                        # This is not a neat way to do this but other methods involve placing wrappers around
                        # anything that sends queries to splitgraph_meta or audit schemas.
                        if _AUDIT_SCHEMA + "." in str(e):
                            raise EngineInitializationError(
                                "Audit triggers not found on the engine. Has the engine been initialized?"
                            ) from e
                        for meta_table in META_TABLES:
                            if "splitgraph_meta.%s" % meta_table in str(e):
                                raise EngineInitializationError(
                                    "splitgraph_meta not found on the engine. Has the engine been initialized?"
                                ) from e
                        if "splitgraph_meta" in str(e):
                            raise ObjectNotFoundError(e)
                    elif isinstance(e, InvalidSchemaName):
                        if "splitgraph_api" in str(e):
                            raise EngineInitializationError(
                                "splitgraph_api not found on the engine. Has the engine been initialized?"
                            ) from e
                    elif (
                        isinstance(e, psycopg2.OperationalError)
                        and (
                            "connection has been closed unexpectedly" in str(e)
                            or "server closed the connection unexpectedly" in str(e)
                        )
                        and attempt == 0
                    ):
                        # Handle the registry closing the connection by retrying once.
                        attempt += 1
                        logging.info(f"Connection to {self.name} lost. Reconnecting...")
                        self.close()
                        connection = self.connection
                        continue

                    raise

                if cur.description is None:
                    return None

                if return_shape == ResultShape.ONE_ONE:
                    result = cur.fetchone()
                    return result[0] if result else None
                if return_shape == ResultShape.ONE_MANY:
                    return cur.fetchone()
                if return_shape == ResultShape.MANY_ONE:
                    return [c[0] for c in cur.fetchall()]
                return cur.fetchall()

    def get_primary_keys(self, schema: str, table: str) -> List[Tuple[str, str]]:
        """Inspects the Postgres information_schema to get the primary keys for a given table."""
        return cast(
            List[Tuple[str, str]],
            self.run_sql(
                """SELECT c.column_name, c.data_type
FROM information_schema.table_constraints tc 
JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name) 
JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema
  AND tc.table_name = c.table_name AND ccu.column_name = c.column_name
WHERE constraint_type = 'PRIMARY KEY'
AND tc.constraint_schema = %s
AND tc.table_name = %s
""",
                (schema, table),
                return_shape=ResultShape.MANY_MANY,
            ),
        )

    def run_sql_batch(
        self,
        statement: Union[Composed, str],
        arguments: Any,
        schema: Optional[str] = None,
        max_size=API_MAX_QUERY_LENGTH,
    ) -> None:
        if not arguments:
            return
        with self.connection.cursor() as cur:
            try:
                if schema:
                    cur.execute("SET search_path to %s;", (schema,))

                batches = _paginate_by_size(
                    cur,
                    statement,
                    arguments,
                    max_size=API_MAX_QUERY_LENGTH,
                )
                for batch in batches:
                    cur.execute(batch)
                if schema:
                    cur.execute("SET search_path to public")
            except DatabaseError:
                self.rollback()
                raise

    def run_api_call(self, call: str, *args, schema: str = SPLITGRAPH_API_SCHEMA) -> Any:
        # When we're inside of a foreign data wrapper on the engine itself,
        # we get to avoid having to go through PostgreSQL to manage objects in
        # the engine's filesystem and call various Python procedures directly.
        # This also avoids the overhead of importing Python modules _again_
        # inside of API plpython funcs.
        if self.in_fdw:
            func = getattr(server, call)
            return func(*args)

        return self.run_sql(
            SQL("SELECT {}.{}" + "(" + ",".join(["%s"] * len(args)) + ")").format(
                Identifier(schema), Identifier(call)
            ),
            args,
            return_shape=ResultShape.ONE_ONE,
        )

    def run_api_call_batch(self, call: str, argslist, schema: str = SPLITGRAPH_API_SCHEMA):
        if self.in_fdw:
            func = getattr(server, call)
            for args in argslist:
                func(*args)
        else:
            self.run_sql_batch(
                SQL("SELECT {}.{}" + "(" + ",".join(["%s"] * len(argslist[0])) + ")").format(
                    Identifier(schema), Identifier(call)
                ),
                argslist,
            )

    def dump_table_sql(
        self,
        schema: str,
        table_name: str,
        stream: TextIOWrapper,
        columns: str = "*",
        where: str = "",
        where_args: Optional[Union[List[str], Tuple[str, str]]] = None,
        target_schema: Optional[str] = None,
        target_table: Optional[str] = None,
    ) -> None:
        target_schema = target_schema or schema
        target_table = target_table or table_name

        with self.connection.cursor() as cur:
            # Don't deserialize JSON into Python dicts (store it as a string)
            psycopg2.extras.register_default_json(cur, globally=False, loads=lambda x: x)
            psycopg2.extras.register_default_jsonb(cur, globally=False, loads=lambda x: x)
            cur.execute(select(table_name, columns, where, schema), where_args)
            if cur.rowcount == 0:
                return

            stream.write(
                SQL("INSERT INTO {}.{} VALUES \n")
                .format(Identifier(target_schema), Identifier(target_table))
                .as_string(self.connection)
            )
            stream.write(
                ",\n".join(
                    cur.mogrify("(" + ",".join(itertools.repeat("%s", len(row))) + ")", row).decode(
                        "utf-8"
                    )
                    for row in cur
                )
            )
        stream.write(";\n")

    def _admin_conn(self) -> "Connection":
        retries = 0
        failed = False

        def _notify():
            nonlocal failed
            print(("\033[F" if failed else "") + "Waiting for connection..." + "." * retries)
            sys.stdout.flush()
            failed = True

        while True:
            try:
                return psycopg2.connect(
                    dbname=self.conn_params["SG_ENGINE_POSTGRES_DB_NAME"],
                    user=self.conn_params.get("SG_ENGINE_ADMIN_USER")
                    or self.conn_params["SG_ENGINE_USER"],
                    password=self.conn_params.get("SG_ENGINE_ADMIN_PWD")
                    or self.conn_params["SG_ENGINE_PWD"],
                    host=self.conn_params["SG_ENGINE_HOST"],
                    port=self.conn_params["SG_ENGINE_PORT"],
                    application_name="sgr " + __version__,
                )
            except psycopg2.errors.DatabaseError as e:
                if retries >= RETRY_AMOUNT:
                    logging.exception(
                        "Error connecting to the engine after %d retries", RETRY_AMOUNT
                    )
                    raise
                _handle_fatal(e)
                retries += 1
                if _quiet():
                    _notify()
                else:
                    logging.error(
                        "Error connecting to the engine (%s: %s), "
                        "sleeping %.2fs and retrying (%d/%d)...",
                        str(e.__class__.__name__),
                        e,
                        RETRY_DELAY,
                        retries,
                        RETRY_AMOUNT,
                    )
                time.sleep(RETRY_DELAY)

    def initialize(
        self, skip_object_handling: bool = False, skip_create_database: bool = False
    ) -> None:
        """Create the Splitgraph Postgres database and install the audit trigger

        :param skip_object_handling: If True, skips installation of
            audit triggers and other object management routines for engines
            that don't need change tracking or checkouts.
        :param skip_create_database: Don't create the Splitgraph database"""

        logging.info("Initializing engine %r...", self)

        # Use the connection to the "postgres" database to create the actual PG_DB
        if skip_create_database:
            logging.info("Explicitly skipping create database")
        else:
            with self._admin_conn() as admin_conn:
                # CREATE DATABASE can't run inside of tx
                pg_db = self.conn_params["SG_ENGINE_DB_NAME"]

                admin_conn.autocommit = True
                with admin_conn.cursor() as cur:
                    cur.execute("SELECT 1 FROM pg_database WHERE datname = %s", (pg_db,))
                    if cur.fetchone() is None:
                        logging.info("Creating database %s", pg_db)
                        cur.execute("ROLLBACK")
                        cur.execute(SQL("CREATE DATABASE {}").format(Identifier(pg_db)))
                    else:
                        logging.info("Database %s already exists, skipping", pg_db)

        logging.info("Ensuring the metadata schema at %s exists...", SPLITGRAPH_META_SCHEMA)

        ensure_metadata_schema(self)

        # Install the push/pull API functions
        logging.info("Installing Splitgraph API functions...")
        push_pull = importlib.resources.read_text(static, _PUSH_PULL)
        self.run_sql(push_pull)

        if skip_object_handling:
            logging.info("Skipping installation of audit triggers/CStore as specified.")
        else:
            # Install CStore management routines
            logging.info("Installing CStore management functions...")
            push_pull = importlib.resources.read_text(static, _CSTORE)
            self.run_sql(push_pull)

            # Install the audit trigger if it doesn't exist
            if not self.schema_exists(_AUDIT_SCHEMA):
                logging.info("Installing the audit trigger...")
                audit_trigger = importlib.resources.read_text(static, _AUDIT_TRIGGER)
                self.run_sql(audit_trigger)
            else:
                logging.info("Skipping the audit trigger as it's already installed.")

            # Start up the pgcrypto extension (required for hashing fragments)
            self.run_sql("CREATE EXTENSION IF NOT EXISTS pgcrypto")

    def delete_database(self, database: str) -> None:
        """
        Helper function to drop a database using the admin connection

        :param database: Database name to drop
        """
        with self._admin_conn() as admin_conn:
            admin_conn.autocommit = True
            with admin_conn.cursor() as cur:
                # For some reason without it we get "DROP DATABASE cannot run inside a transaction block"
                cur.execute("ROLLBACK")
                cur.execute(SQL("DROP DATABASE IF EXISTS {}").format(Identifier(database)))


class AuditTriggerChangeEngine(PsycopgEngine, ChangeEngine):
    """Change tracking based on an audit trigger stored procedure"""

    def get_tracked_tables(self) -> List[Tuple[str, str]]:
        """Return a list of tables that the audit trigger is working on."""
        return cast(
            List[Tuple[str, str]],
            self.run_sql(
                "SELECT DISTINCT event_object_schema, event_object_table "
                "FROM information_schema.triggers WHERE trigger_name IN (%s, %s)",
                (ROW_TRIGGER_NAME, STM_TRIGGER_NAME),
            ),
        )

    def track_tables(self, tables: List[Tuple[str, str]]) -> None:
        """Install the audit trigger on the required tables"""
        self.run_sql(
            SQL(";").join(
                itertools.repeat(
                    SQL("SELECT {}.audit_table(%s)").format(Identifier(_AUDIT_SCHEMA)), len(tables)
                )
            ),
            ["{}.{}".format(_quote_ident(s), _quote_ident(t)) for s, t in tables],
        )

    def untrack_tables(self, tables: List[Tuple[str, str]]) -> None:
        """Remove triggers from tables and delete their pending changes"""
        for trigger in (ROW_TRIGGER_NAME, STM_TRIGGER_NAME):
            self.run_sql(
                SQL(";").join(
                    SQL("DROP TRIGGER IF EXISTS {} ON {}.{}").format(
                        Identifier(trigger), Identifier(s), Identifier(t)
                    )
                    for s, t in tables
                )
            )
        # Delete the actual logged actions for untracked tables
        self.run_sql_batch(
            SQL("DELETE FROM {}.logged_actions WHERE schema_name = %s AND table_name = %s").format(
                Identifier(_AUDIT_SCHEMA)
            ),
            tables,
        )

    def has_pending_changes(self, schema: str) -> bool:
        """
        Return True if the tracked schema has pending changes and False if it doesn't.
        """
        return (
            self.run_sql(
                SQL("SELECT 1 FROM {}.{} WHERE schema_name = %s").format(
                    Identifier(_AUDIT_SCHEMA), Identifier("logged_actions")
                ),
                (schema,),
                return_shape=ResultShape.ONE_ONE,
            )
            is not None
        )

    def discard_pending_changes(self, schema: str, table: Optional[str] = None) -> None:
        """
        Discard recorded pending changes for a tracked schema / table
        """
        query = SQL("DELETE FROM {}.{} WHERE schema_name = %s").format(
            Identifier(_AUDIT_SCHEMA), Identifier("logged_actions")
        )

        if table:
            self.run_sql(
                query + SQL(" AND table_name = %s"), (schema, table), return_shape=ResultShape.NONE
            )
        else:
            self.run_sql(query, (schema,), return_shape=ResultShape.NONE)

    def get_pending_changes(
        self, schema: str, table: str, aggregate: bool = False
    ) -> Union[
        List[Tuple[int, int]], List[Tuple[Tuple[str, ...], bool, Dict[str, Any], Dict[str, Any]]]
    ]:
        """
        Return pending changes for a given tracked table

        :param schema: Schema the table belongs to
        :param table: Table to return changes for
        :param aggregate: Whether to aggregate changes or return them completely
        :return: If aggregate is True: List of tuples of (change_type, number of rows).
            If aggregate is False: List of (primary_key, change_type, change_data)
        """
        if aggregate:
            return [
                (_KIND[k], c)
                for k, c in self.run_sql(
                    SQL(
                        "SELECT action, count(action) FROM {}.{} "
                        "WHERE schema_name = %s AND table_name = %s GROUP BY action"
                    ).format(Identifier(_AUDIT_SCHEMA), Identifier("logged_actions")),
                    (schema, table),
                )
            ]

        ri_cols, _ = zip(*self.get_change_key(schema, table))
        result: List[Tuple[Tuple, bool, Dict, Dict]] = []
        for action, row_data, changed_fields in self.run_sql(
            SQL(
                "SELECT action, row_data, changed_fields FROM {}.{} "
                "WHERE schema_name = %s AND table_name = %s"
            ).format(Identifier(_AUDIT_SCHEMA), Identifier("logged_actions")),
            (schema, table),
        ):
            result.extend(_convert_audit_change(action, row_data, changed_fields, ri_cols))
        return result

    def get_changed_tables(self, schema: str) -> List[str]:
        """Get list of tables that have changed content"""
        return cast(
            List[str],
            self.run_sql(
                SQL(
                    """SELECT DISTINCT(table_name) FROM {}.{}
                               WHERE schema_name = %s"""
                ).format(Identifier(_AUDIT_SCHEMA), Identifier("logged_actions")),
                (schema,),
                return_shape=ResultShape.MANY_ONE,
            ),
        )


class PostgresEngine(AuditTriggerChangeEngine, ObjectEngine):
    """An implementation of the Postgres engine for Splitgraph"""

    def get_object_schema(self, object_id: str) -> "TableSchema":
        result: "TableSchema" = []

        call_result = self.run_api_call("get_object_schema", object_id)
        if isinstance(call_result, str):
            call_result = json.loads(call_result)

        for ordinal, column_name, column_type, is_pk in call_result:
            assert isinstance(ordinal, int)
            assert isinstance(column_name, str)
            assert isinstance(column_type, str)
            assert isinstance(is_pk, bool)
            result.append(TableColumn(ordinal, column_name, column_type, is_pk))

        return result

    def _set_object_schema(self, object_id: str, schema_spec: "TableSchema") -> None:
        # Drop the comments from the schema spec (not stored in the object schema file).
        schema_spec = [s[:4] for s in schema_spec]
        self.run_api_call("set_object_schema", object_id, json.dumps(schema_spec))

    def dump_object_creation(
        self,
        object_id: str,
        schema: str,
        table: Optional[str] = None,
        schema_spec: Optional["TableSchema"] = None,
        if_not_exists: bool = False,
    ) -> bytes:
        """
        Generate the SQL that remounts a foreign table pointing to a Splitgraph object.

        :param object_id: Name of the object
        :param schema: Schema to create the table in
        :param table: Name of the table to mount
        :param schema_spec: Schema of the table
        :param if_not_exists: Add IF NOT EXISTS to the DDL
        :return: SQL in bytes format.
        """
        table = table or object_id

        if not schema_spec:
            schema_spec = self.get_object_schema(object_id)
        query = SQL(
            "CREATE FOREIGN TABLE " + ("IF NOT EXISTS " if if_not_exists else "") + "{}.{} ("
        ).format(Identifier(schema), Identifier(table))
        query += SQL(",".join("{} %s " % validate_type(col.pg_type) for col in schema_spec)).format(
            *(Identifier(col.name) for col in schema_spec)
        )
        # foreign tables/cstore don't support PKs
        query += SQL(") SERVER {} OPTIONS (compression %s, filename %s)").format(
            Identifier(CSTORE_SERVER)
        )

        object_path = self.conn_params["SG_ENGINE_OBJECT_PATH"]
        assert object_path is not None

        with self.connection.cursor() as cur:
            return cast(
                bytes, cur.mogrify(query, ("pglz", str(PurePosixPath(object_path, object_id))))
            )

    def dump_object(self, object_id: str, stream: TextIOWrapper, schema: str) -> None:
        schema_spec = self.get_object_schema(object_id)
        stream.write(
            self.dump_object_creation(object_id, schema=schema, schema_spec=schema_spec).decode(
                "utf-8"
            )
        )
        stream.write(";\n")
        with self.connection.cursor() as cur:
            # Since we can't write into the CStore table directly, we first load the data
            # into a temporary table and then insert that data into the CStore table.
            query, args = self.dump_table_creation(
                None, "cstore_tmp_ingestion", schema_spec, temporary=True
            )
            stream.write(cur.mogrify(query, args).decode("utf-8"))
            stream.write(";\n")
            self.dump_table_sql(
                schema,
                object_id,
                stream,
                target_schema="pg_temp",
                target_table="cstore_tmp_ingestion",
            )
            stream.write(
                SQL("INSERT INTO {}.{} (SELECT * FROM pg_temp.cstore_tmp_ingestion);\n")
                .format(Identifier(schema), Identifier(object_id))
                .as_string(cur)
            )
            stream.write(
                cur.mogrify(
                    "SELECT splitgraph_api.set_object_schema(%s, %s);\n",
                    (object_id, json.dumps(schema_spec)),
                ).decode("utf-8")
            )
            stream.write("DROP TABLE pg_temp.cstore_tmp_ingestion;\n")

    def get_object_size(self, object_id: str) -> int:
        return int(self.run_api_call("get_object_size", object_id))

    def delete_objects(self, object_ids: List[str]) -> None:
        self.unmount_objects(object_ids)
        self.run_api_call_batch("delete_object_files", [(o,) for o in object_ids])

    def rename_object(self, old_object_id: str, new_object_id: str):
        self.unmount_objects([old_object_id])
        self.run_api_call("rename_object_files", old_object_id, new_object_id)
        self.mount_object(new_object_id)

    def unmount_objects(self, object_ids: List[str]) -> None:
        """Unmount objects from splitgraph_meta (this doesn't delete the physical files."""
        unmount_query = SQL(";").join(
            SQL("DROP FOREIGN TABLE IF EXISTS {}.{}").format(
                Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
            )
            for object_id in object_ids
        )
        self.run_sql(unmount_query)

    def sync_object_mounts(self) -> None:
        """Scan through local object storage and synchronize it with the foreign tables in
        splitgraph_meta (unmounting non-existing objects and mounting existing ones)."""
        object_ids = self.run_api_call("list_objects")

        mounted_objects = self.run_sql(
            "SELECT table_name FROM information_schema.tables "
            "WHERE table_schema = %s AND table_type = 'FOREIGN'",
            (SPLITGRAPH_META_SCHEMA,),
            return_shape=ResultShape.MANY_ONE,
        )

        if mounted_objects:
            self.unmount_objects(mounted_objects)

        for object_id in object_ids:
            self.mount_object(object_id, schema_spec=self.get_object_schema(object_id))

    def mount_object(
        self,
        object_id: str,
        table: None = None,
        schema: str = SPLITGRAPH_META_SCHEMA,
        schema_spec: Optional["TableSchema"] = None,
    ) -> None:
        """
        Mount an object from local storage as a foreign table.

        :param object_id: ID of the object
        :param table: Table to mount the object into
        :param schema: Schema to mount the object into
        :param schema_spec: Schema of the object.
        """
        query = self.dump_object_creation(object_id, schema, table, schema_spec, if_not_exists=True)
        try:
            self.run_sql(query)
        except psycopg2.errors.UndefinedObject as e:
            raise ObjectMountingError(
                "Error mounting. It's possible that the object was created "
                "using an extension (e.g. PostGIS) that's not installed on this engine."
            ) from e

    def store_fragment(
        self,
        inserted: Any,
        deleted: Any,
        schema: str,
        table: str,
        source_schema: str,
        source_table: str,
        source_schema_spec: Optional[TableSchema] = None,
    ) -> None:
        temporary = schema == "pg_temp"

        schema_spec = source_schema_spec or self.get_full_table_schema(source_schema, source_table)

        # Assuming the schema_spec has the whole tuple as PK if the table has no PK.
        change_key = get_change_key(schema_spec)
        ri_cols, ri_types = zip(*change_key)
        ri_cols = list(ri_cols)
        ri_types = list(ri_types)
        non_ri_cols = [c.name for c in schema_spec if c.name not in ri_cols]
        all_cols = ri_cols + non_ri_cols

        self.create_table(
            schema,
            table,
            schema_spec=add_ud_flag_column(schema_spec),
            temporary=temporary,
        )

        # Store upserts
        # INSERT INTO target_table (sg_ud_flag, col1, col2...)
        #   (SELECT true, t.col1, t.col2, ...
        #    FROM VALUES ((pk1_1, pk1_2), (pk2_1, pk2_2)...)) v JOIN source_table t
        #    ON v.pk1 = t.pk1::pk1_type AND v.pk2::pk2_type = t.pk2...
        #    -- the cast is required since the audit trigger gives us strings for values of updated columns
        #    -- and we're intending to join those with the PKs in the original table.
        if inserted:
            if non_ri_cols:
                query = (
                    SQL("INSERT INTO {}.{} (").format(Identifier(schema), Identifier(table))
                    + SQL(",").join(Identifier(c) for c in [SG_UD_FLAG] + all_cols)
                    + SQL(")")
                    + SQL("(SELECT %s, ")
                    + SQL(",").join(SQL("t.") + Identifier(c) for c in all_cols)
                    + SQL(
                        " FROM (VALUES "
                        + ",".join(
                            itertools.repeat(
                                "(" + ",".join(itertools.repeat("%s", len(inserted[0]))) + ")",
                                len(inserted),
                            )
                        )
                        + ")"
                    )
                    + SQL(" AS v (")
                    + SQL(",").join(Identifier(c) for c in ri_cols)
                    + SQL(")")
                    + SQL("JOIN {}.{} t").format(
                        Identifier(source_schema), Identifier(source_table)
                    )
                    + SQL(" ON ")
                    + SQL(" AND ").join(
                        SQL("t.{0} = v.{0}::%s" % r).format(Identifier(c))
                        for c, r in zip(ri_cols, ri_types)
                    )
                    + SQL(")")
                )
                # Flatten the args
                args = [True] + [p for pk in inserted for p in pk]
            else:
                # If the whole tuple is the PK, there's no point joining on the actual source table
                query = (
                    SQL("INSERT INTO {}.{} (").format(Identifier(schema), Identifier(table))
                    + SQL(",").join(Identifier(c) for c in [SG_UD_FLAG] + ri_cols)
                    + SQL(")")
                    + SQL(
                        "VALUES "
                        + ",".join(
                            itertools.repeat(
                                "(" + ",".join(itertools.repeat("%s", len(inserted[0]) + 1)) + ")",
                                len(inserted),
                            )
                        )
                    )
                )
                args = [p for pk in inserted for p in [True] + list(pk)]
            self.run_sql(query, args)

        # Store the deletes
        # we don't actually have the old values here so we put NULLs (which should be compressed out).
        if deleted:
            query = (
                SQL("INSERT INTO {}.{} (").format(Identifier(schema), Identifier(table))
                + SQL(",").join(Identifier(c) for c in [SG_UD_FLAG] + ri_cols)
                + SQL(")")
                + SQL(
                    "VALUES "
                    + ",".join(
                        itertools.repeat(
                            "(" + ",".join(itertools.repeat("%s", len(deleted[0]) + 1)) + ")",
                            len(deleted),
                        )
                    )
                )
            )
            args = [p for pk in deleted for p in [False] + list(pk)]
            self.run_sql(query, args)

    def store_object(
        self,
        object_id: str,
        source_query: Union[bytes, Composed, str, SQL],
        schema_spec: TableSchema,
        source_query_args=None,
        overwrite=False,
    ) -> None:

        # The physical object storage (/var/lib/splitgraph/objects) and the actual
        # foreign tables in spligraph_meta might be desynced for a variety of reasons,
        # so we have to handle all four cases of (physical object exists/doesn't exist,
        # foreign table exists/doesn't exist).
        object_exists = bool(self.run_api_call("object_exists", object_id))

        # Try mounting the object
        if self.table_exists(SPLITGRAPH_META_SCHEMA, object_id):
            # Foreign table with that name already exists. If it does exist but there's no
            # physical object file, we'll have to write into the table anyway.
            if not object_exists:
                logging.info(
                    "Object storage, %s, mounted but no physical file, recreating",
                    object_id,
                )

            # In addition, there's a corner case where the object was mounted with different FDW
            # parameters (e.g. object path in /var/lib/splitgraph/objects has changed)
            # and we want that to be overwritten in any case, so we remount the object.
            self.unmount_objects([object_id])
            self.mount_object(object_id, schema_spec=schema_spec)
        else:
            self.mount_object(object_id, schema_spec=schema_spec)

        if object_exists:
            if overwrite:
                logging.info("Object %s already exists, will overwrite", object_id)
                self.run_sql(
                    SQL("TRUNCATE TABLE {}.{}").format(
                        Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
                    )
                )
            else:
                logging.info("Object %s already exists, skipping", object_id)
                return

        # At this point, the foreign table mounting the object exists and we've established
        # that it's a brand new table, so insert data into it.
        self.run_sql(
            SQL("INSERT INTO {}.{}").format(
                Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
            )
            + source_query,
            source_query_args,
        )

        # Also store the table schema in a file
        self._set_object_schema(object_id, schema_spec)

    @staticmethod
    def _schema_spec_to_cols(schema_spec: "TableSchema") -> Tuple[List[str], List[str]]:
        pk_cols = [p.name for p in schema_spec if p.is_pk]
        non_pk_cols = [p.name for p in schema_spec if not p.is_pk]

        if not pk_cols:
            pk_cols = [p.name for p in schema_spec if p.pg_type in PG_INDEXABLE_TYPES]
            non_pk_cols = [p.name for p in schema_spec if p.pg_type not in PG_INDEXABLE_TYPES]
        return pk_cols, non_pk_cols

    @staticmethod
    def _generate_fragment_application(
        source_schema: str,
        source_table: str,
        target_schema: str,
        target_table: str,
        cols: Tuple[List[str], List[str]],
        extra_quals: Optional[Composed] = None,
    ) -> Composed:
        ri_cols, non_ri_cols = cols
        all_cols = ri_cols + non_ri_cols

        # First, delete all PKs from staging that are mentioned in the new fragment. This conveniently
        # covers both deletes and updates.

        # Also, alias tables so that we don't have to send around long strings of object IDs.
        query = (
            SQL("DELETE FROM {0}.{2} t USING {1}.{3} s").format(
                Identifier(target_schema),
                Identifier(source_schema),
                Identifier(target_table),
                Identifier(source_table),
            )
            + SQL(" WHERE ")
            + _generate_where_clause("t", ri_cols, "s")
        )

        # At this point, we can insert all rows directly since we won't have any conflicts.
        # We can also apply extra qualifiers to only insert rows that match a certain query,
        # which will result in fewer rows actually being written to the staging table.

        # INSERT INTO target_table (col1, col2...)
        #   (SELECT col1, col2, ...
        #    FROM fragment_table WHERE sg_ud_flag = true (AND optional quals))
        query += (
            SQL(";INSERT INTO {}.{} (").format(Identifier(target_schema), Identifier(target_table))
            + SQL(",").join(Identifier(c) for c in all_cols)
            + SQL(")")
            + SQL("(SELECT ")
            + SQL(",").join(Identifier(c) for c in all_cols)
            + SQL(" FROM {}.{}").format(Identifier(source_schema), Identifier(source_table))
        )
        extra_quals = [extra_quals] if extra_quals else []
        extra_quals.append(SQL("{} = true").format(Identifier(SG_UD_FLAG)))
        if extra_quals:
            query += SQL(" WHERE ") + SQL(" AND ").join(extra_quals)
        query += SQL(")")
        return query

    def apply_fragments(
        self,
        objects: List[Tuple[str, str]],
        target_schema: str,
        target_table: str,
        extra_quals: Optional[Composed] = None,
        extra_qual_args: Optional[Tuple[Any, ...]] = None,
        schema_spec: Optional["TableSchema"] = None,
        progress_every: Optional[int] = None,
    ) -> None:
        if not objects:
            return
        schema_spec = schema_spec or self.get_full_table_schema(target_schema, target_table)
        # Assume that the target table already has the required schema (including PKs)
        # and use that to generate queries to apply fragments.
        cols = self._schema_spec_to_cols(schema_spec)

        if progress_every:
            batches = list(chunk(objects, chunk_size=progress_every))
            with tqdm(total=len(objects), unit="obj") as pbar:
                for batch in batches:
                    self._apply_batch(
                        batch, target_schema, target_table, extra_quals, cols, extra_qual_args
                    )
                    pbar.update(len(batch))
                    pbar.set_postfix({"object": batch[-1][1][:10] + "..."})
        else:
            self._apply_batch(
                objects, target_schema, target_table, extra_quals, cols, extra_qual_args
            )

    def _apply_batch(
        self, objects, target_schema, target_table, extra_quals, cols, extra_qual_args
    ):
        query = SQL(";").join(
            self._generate_fragment_application(
                ss, st, target_schema, target_table, cols, extra_quals
            )
            for ss, st in objects
        )
        self.run_sql(query, (extra_qual_args * len(objects)) if extra_qual_args else None)

    def upload_objects(self, objects: List[str], remote_engine: "PostgresEngine") -> None:

        # We don't have direct access to the remote engine's storage and we also
        # can't use the old method of first creating a CStore table remotely and then
        # mounting it via FDW (because this is done on two separate connections, after
        # the table is created and committed on the first one, it can't be written into.
        #
        # So we have to do a slower method of dumping the table into binary
        # and then piping that binary into the remote engine.
        #
        # Perhaps we should drop direct uploading altogether and require people to use S3 throughout.

        pbar = tqdm(objects, unit="objs", ascii=SG_CMD_ASCII)
        for object_id in pbar:
            pbar.set_postfix(object=object_id[:10] + "...")
            schema_spec = self.get_object_schema(object_id)
            remote_engine.mount_object(object_id, schema_spec=schema_spec)

            # Truncate the remote object in case it already exists (we'll overwrite it).
            remote_engine.run_sql(
                SQL("TRUNCATE TABLE {}.{}").format(
                    Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
                )
            )

            stream = BytesIO()
            with self.copy_cursor() as cur:
                cur.copy_expert(
                    SQL("COPY {}.{} TO STDOUT WITH (FORMAT 'binary')").format(
                        Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
                    ),
                    stream,
                )
            stream.seek(0)
            with remote_engine.copy_cursor() as cur:
                cur.copy_expert(
                    SQL("COPY {}.{} FROM STDIN WITH (FORMAT 'binary')").format(
                        Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
                    ),
                    stream,
                )
            remote_engine._set_object_schema(object_id, schema_spec)
            remote_engine.commit()

    @contextmanager
    def _mount_remote_engine(self, remote_engine: "PostgresEngine") -> Iterator[str]:
        # Switch the global engine to "self" instead of LOCAL since `mount_postgres` uses the global engine
        # (we can't easily pass args into it as it's also invoked from the command line with its arguments
        # used to populate --help)
        user = remote_engine.conn_params["SG_ENGINE_USER"]
        pwd = remote_engine.conn_params["SG_ENGINE_PWD"]
        host = remote_engine.conn_params["SG_ENGINE_FDW_HOST"]
        port = remote_engine.conn_params["SG_ENGINE_FDW_PORT"]
        dbname = remote_engine.conn_params["SG_ENGINE_DB_NAME"]

        # conn_params might contain Nones for some parameters (host/port for Unix
        # socket connection, so here we have to validate that they aren't None).
        assert user is not None
        assert pwd is not None
        assert host is not None
        assert port is not None
        assert dbname is not None

        logging.info(
            "Mounting remote schema %s@%s:%s/%s/%s to %s...",
            user,
            host,
            port,
            dbname,
            SPLITGRAPH_META_SCHEMA,
            REMOTE_TMP_SCHEMA,
        )
        self.delete_schema(REMOTE_TMP_SCHEMA)
        with switch_engine(self):
            mount_postgres(
                mountpoint=REMOTE_TMP_SCHEMA,
                host=host,
                port=int(port),
                username=user,
                password=pwd,
                dbname=dbname,
                remote_schema=SPLITGRAPH_META_SCHEMA,
            )
        try:
            yield REMOTE_TMP_SCHEMA
        finally:
            self.delete_schema(REMOTE_TMP_SCHEMA)

    def download_objects(self, objects: List[str], remote_engine: "PostgresEngine") -> List[str]:
        # Instead of connecting and pushing queries to it from the Python client, we just mount the remote mountpoint
        # into a temporary space (without any checking out) and SELECT the  required data into our local tables.

        with self._mount_remote_engine(remote_engine) as remote_schema:
            downloaded_objects = []
            pbar = tqdm(objects, unit="objs", ascii=SG_CMD_ASCII)
            for object_id in pbar:
                pbar.set_postfix(object=object_id[:10] + "...")
                if not self.table_exists(remote_schema, object_id):
                    logging.error("%s not found on the remote engine!", object_id)
                    continue

                # Create the CStore table on the engine and copy the contents of the object into it.
                schema_spec = remote_engine.get_full_table_schema(SPLITGRAPH_META_SCHEMA, object_id)
                self.mount_object(object_id, schema_spec=schema_spec)
                self.copy_table(
                    remote_schema,
                    object_id,
                    SPLITGRAPH_META_SCHEMA,
                    object_id,
                    with_pk_constraints=False,
                )
                self._set_object_schema(object_id, schema_spec=schema_spec)
                downloaded_objects.append(object_id)
        if len(downloaded_objects) < len(objects):
            raise IncompleteObjectDownloadError(reason=None, successful_objects=downloaded_objects)
        return downloaded_objects

    def get_change_key(self, schema: str, table: str) -> List[Tuple[str, str]]:
        return get_change_key(self.get_full_table_schema(schema, table))


def get_change_key(schema_spec: TableSchema) -> List[Tuple[str, str]]:
    pk = [(c.name, c.pg_type) for c in schema_spec if c.is_pk]
    if pk:
        return pk

    # Only return columns that can be compared (since we'll be using them
    # for chunking)
    return [(c.name, c.pg_type) for c in schema_spec if c.pg_type in PG_INDEXABLE_TYPES]


def _split_ri_cols(
    action: str,
    row_data: Dict[str, Any],
    changed_fields: Optional[Dict[str, str]],
    ri_cols: Tuple[str, ...],
) -> Any:
    """
    :return: `(ri_data, non_ri_data)`: a tuple of 2 dictionaries:
        * `ri_data`: maps column names in `ri_cols` to values identifying the replica identity (RI) of a given tuple
        * `non_ri_data`: map of column names and values not in the RI that have been changed/updated
    """
    non_ri_data = {}
    ri_data = {}

    if action == "I":
        for column, value in row_data.items():
            if column in ri_cols:
                ri_data[column] = value
            else:
                non_ri_data[column] = value
    elif action == "D":
        for column, value in row_data.items():
            if column in ri_cols:
                ri_data[column] = value
    else:
        assert action == "U"
        for column, value in row_data.items():
            if column in ri_cols:
                ri_data[column] = value
        if changed_fields:
            for column, value in changed_fields.items():
                non_ri_data[column] = value

    return ri_data, non_ri_data


def _recalculate_disjoint_ri_cols(
    ri_cols: Tuple[str, ...],
    ri_data: Dict[str, Any],
    non_ri_data: Dict[str, Any],
    row_data: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    # If part of the PK has been updated (is in the non_ri_cols/vals), we have to instead
    # apply the update to the PK (ri_cols/vals) and recalculate the new full new tuple
    # (by applying the update to row_data).
    new_non_ri_data = {}
    row_data = row_data.copy()

    for nrc, nrv in non_ri_data.items():
        if nrc in ri_cols:
            ri_data[nrc] = nrv
        else:
            row_data[nrc] = nrv

    for col, val in row_data.items():
        if col not in ri_cols:
            new_non_ri_data[col] = val

    return ri_data, new_non_ri_data


def _convert_audit_change(
    action: str,
    row_data: Dict[str, Any],
    changed_fields: Optional[Dict[str, str]],
    ri_cols: Tuple[str, ...],
) -> List[Tuple[Tuple, bool, Dict[str, str], Dict[str, str]]]:
    """
    Converts the audit log entry into Splitgraph's internal format.

    :returns: [(pk, (True for upserted, False for deleted), (old row value if updated/deleted),
        (new row value if inserted/updated))].
        More than 1 change might be emitted from a single audit entry.
    """
    ri_data, non_ri_data = _split_ri_cols(action, row_data, changed_fields, ri_cols)
    pk_changed = any(c in ri_cols for c in non_ri_data)

    if changed_fields:
        new_row = row_data.copy()
        for key, value in changed_fields.items():
            if key not in ri_cols:
                new_row[key] = value
    else:
        new_row = row_data
    if pk_changed:
        assert action == "U"
        # If it's an update that changed the PK (e.g. the table has no replica identity so we treat the whole
        # tuple as a primary key), then we turn it into (old PK, DELETE, old row data); (new PK, INSERT, {})

        # Recalculate the new PK to be inserted + the new (full) tuple, otherwise if the whole
        # tuple hasn't been updated, we'll lose parts of the old row (see test_diff_conflation_on_commit[test_case2]).

        result = [(tuple(ri_data[c] for c in ri_cols), False, row_data, new_row)]

        ri_data, non_ri_data = _recalculate_disjoint_ri_cols(
            ri_cols, ri_data, non_ri_data, row_data
        )
        result.append((tuple(ri_data[c] for c in ri_cols), True, {}, new_row))
        return result
    if action == "U" and not non_ri_data:
        # Nothing was actually updated -- don't emit an action
        return []
    return [
        (
            tuple(ri_data[c] for c in ri_cols),
            action in ("I", "U"),
            row_data if action in ("U", "D") else {},
            new_row if action in ("I", "U") else {},
        )
    ]


_KIND = {"I": 0, "D": 1, "U": 2}


def _generate_where_clause(table: str, cols: List[str], table_2: str) -> Composed:
    return SQL(" AND ").join(
        SQL("{}.{} = {}.{}").format(
            Identifier(table), Identifier(c), Identifier(table_2), Identifier(c)
        )
        for c in cols
    )


def add_ud_flag_column(table_schema: TableSchema) -> TableSchema:
    return table_schema + [TableColumn(table_schema[-1].ordinal + 1, SG_UD_FLAG, "boolean", False)]



"""
Extra wrapper code for mount handlers
"""
from typing import TYPE_CHECKING, Any, Dict, Optional

from splitgraph.core.common import unmount_schema
from splitgraph.exceptions import DataSourceError

if TYPE_CHECKING:
    from splitgraph.core.types import TableInfo


def mount_postgres(mountpoint, **kwargs) -> None:
    """
    Mount a Postgres database.

    Mounts a schema on a remote Postgres database as a set of foreign tables locally.
    \b

    :param mountpoint: Schema to mount the remote into.
    :param server: Database hostname.
    :param port: Port the Postgres server is running on.
    :param username: A read-only user that the database will be accessed as.
    :param password: Password for the read-only user.
    :param dbname: Remote database name.
    :param remote_schema: Remote schema name.
    :param extra_server_args: Dictionary of extra arguments to pass to the foreign server
    :param tables: Tables to mount (default all). If a list, then will use IMPORT FOREIGN SCHEMA.
    If a dictionary, must have the format {"table_name": {"col_1": "type_1", ...}}.
    """
    mount(mountpoint, mount_handler="postgres_fdw", handler_kwargs=kwargs, overwrite=True)


def mount(
    mountpoint: str,
    mount_handler: str,
    handler_kwargs: Dict[str, Any],
    overwrite: bool = True,
    tables: Optional["TableInfo"] = None,
) -> None:
    """
    Mounts a foreign database via an FDW (without creating new Splitgraph objects)

    :param mountpoint: Mountpoint to import the new tables into.
    :param mount_handler: The type of the mounted database.
    :param handler_kwargs: Dictionary of options to pass to the mount handler.
    :param overwrite: Delete the foreign server if it already exists. Used by mount_postgres for data pulls.
    :param tables: List of tables to mount or their schemas
    """
    # Workaround for circular imports
    from splitgraph.engine import get_engine
    from splitgraph.hooks.data_source import get_data_source
    from splitgraph.hooks.data_source.fdw import ForeignDataWrapperDataSource

    engine = get_engine()
    data_source = get_data_source(mount_handler)

    if not data_source.supports_mount:
        raise DataSourceError("Data source %s does not support mounting!")

    assert issubclass(data_source, ForeignDataWrapperDataSource)

    unmount_schema(engine, mountpoint)

    source = data_source.from_commandline(engine, handler_kwargs)
    source.mount(schema=mountpoint, overwrite=overwrite, tables=tables)
    engine.commit()



"""
Functions for communicating with the remote Splitgraph catalog
"""

from typing import TYPE_CHECKING, Optional, Union, cast

from psycopg2.sql import SQL, Identifier

from splitgraph.config import SPLITGRAPH_API_SCHEMA, SPLITGRAPH_META_SCHEMA
from splitgraph.engine import ResultShape

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine


def get_info_key(engine: "PostgresEngine", key: str) -> Optional[str]:
    """
    Gets a configuration key from the remote registry, used to notify the client of the registry's capabilities.

    :param engine: Engine
    :param key: Key to get
    """
    return cast(
        Optional[str],
        engine.run_sql(
            SQL("SELECT value FROM {}.info WHERE key = %s").format(
                Identifier(SPLITGRAPH_META_SCHEMA)
            ),
            (key,),
            return_shape=ResultShape.ONE_ONE,
        ),
    )


def set_info_key(engine: "PostgresEngine", key: str, value: Union[bool, str]) -> None:
    """
    Sets a configuration value on the remote registry.

    :param engine: Engine
    :param key: Key to set
    :param value: New value for the key
    """

    engine.run_sql(
        SQL(
            "INSERT INTO {0}.info (key, value) VALUES (%s, %s)"
            " ON CONFLICT (key) DO UPDATE SET value = excluded.value WHERE info.key = excluded.key"
        ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
        (key, value),
    )


def setup_registry_mode(engine: "PostgresEngine") -> None:
    """
    Set up access policies/RLS:

    * Normal users aren't allowed to create tables/schemata (can't do checkouts inside of a registry or
      upload SG objects directly to it)
    * Normal users can't access the splitgraph_meta schema directly: they're only supposed to be able to
      talk to it via stored procedures in splitgraph_api. Those procedures are set up with SECURITY INVOKER
      (run with those users' credentials) and what they can access is further restricted by RLS:

      * images/tables/tags meta tables: can only create/update/delete records where the namespace = user ID
      * objects/object_location tables: same. An object (piece of data) becomes owned by the user that creates
        it and still remains so even if someone else's image starts using it. Hence, the original owner can delete
        or change it (since they control the external location they've uploaded it to anyway).

    """

    if get_info_key(engine, "registry_mode") == "true":
        return

    for schema in (SPLITGRAPH_META_SCHEMA, SPLITGRAPH_API_SCHEMA):
        engine.run_sql(SQL("REVOKE CREATE ON SCHEMA {} FROM PUBLIC").format(Identifier(schema)))
        # Allow schema usage (including splitgraph_meta) but don't actually allow writing to/reading from
        # meta tables. This is so that users can still download objects that are stored directly
        # on the registry.
        engine.run_sql(SQL("GRANT USAGE ON SCHEMA {} TO PUBLIC").format(Identifier(schema)))

    # Allow everyone to read objects that have been uploaded
    engine.run_sql(
        SQL("ALTER DEFAULT PRIVILEGES IN SCHEMA {} GRANT SELECT ON TABLES TO PUBLIC").format(
            Identifier(SPLITGRAPH_META_SCHEMA)
        )
    )

    set_info_key(engine, "registry_mode", "true")



"""Routines that ingest/export CSV files to/from Splitgraph images using Pandas"""

import csv
from io import StringIO
from typing import TYPE_CHECKING, Optional, Union

import pandas as pd
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from pandas.io.sql import get_schema
from psycopg2.sql import SQL, Identifier
from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine

from splitgraph.core.image import Image
from splitgraph.core.repository import Repository
from splitgraph.ingestion.common import IngestionAdapter
from splitgraph.ingestion.csv import copy_csv_buffer

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine, PsycopgEngine


def _get_sqlalchemy_engine(engine: "PostgresEngine") -> Engine:
    server, port, username, password, dbname = (
        engine.conn_params["SG_ENGINE_HOST"],
        engine.conn_params["SG_ENGINE_PORT"],
        engine.conn_params["SG_ENGINE_USER"],
        engine.conn_params["SG_ENGINE_PWD"],
        engine.conn_params["SG_ENGINE_DB_NAME"],
    )
    return create_engine("postgresql://%s:%s@%s:%s/%s" % (username, password, server, port, dbname))


class PandasIngestionAdapter(IngestionAdapter):
    @staticmethod
    def create_ingestion_table(data, engine, schema: str, table: str, **kwargs):
        engine.delete_table(schema, table)
        # Use sqlalchemy's engine to convert types and create a DDL statement for the table.

        # If there's an unnamed index (created by Pandas), we don't add PKs to the table.
        if data.index.names == [None]:
            ddl = get_schema(data, name=table, con=_get_sqlalchemy_engine(engine))
        else:
            ddl = get_schema(
                data.reset_index(),
                name=table,
                keys=data.index.names,
                con=_get_sqlalchemy_engine(engine),
            )
        engine.run_sql_in(schema, ddl)

    @staticmethod
    def data_to_new_table(
        data, engine: "PsycopgEngine", schema: str, table: str, no_header: bool = True, **kwargs
    ):
        df_to_table_fast(engine, data, schema, table)

    @staticmethod
    def query_to_data(engine, query: str, schema: Optional[str] = None, **kwargs):
        # Pandas' `read_sql_table/query` because they has type inference via SQLAlchemy
        # (from the datatypes in the query that postgres gives back).
        if schema:
            query = (
                SQL("SET search_path TO {},public;")
                .format(Identifier(schema))
                .as_string(engine.connection)
                + query
            )

        return pd.read_sql_query(sql=query, con=_get_sqlalchemy_engine(engine), **kwargs)


def df_to_table_fast(
    engine: "PsycopgEngine", df: Union[Series, DataFrame], target_schema: str, target_table: str
):
    # Instead of using Pandas' to_sql, dump the dataframe to csv and then load it on the other
    # end using Psycopg's copy_to.
    # Don't write the index column if it's unnamed (generated by Pandas)
    csv_str = df.to_csv(
        header=False, index=df.index.names != [None], escapechar="\\", quoting=csv.QUOTE_ALL
    )
    # Dirty hack
    csv_str = csv_str.replace('""', "")
    buffer = StringIO()
    buffer.write(csv_str)
    buffer.seek(0)
    copy_csv_buffer(buffer, engine, target_schema, target_table, no_header=True)


_pandas_adapter = PandasIngestionAdapter()


def sql_to_df(
    sql: str,
    image: Optional[Union[Image, str]] = None,
    repository: Optional[Repository] = None,
    use_lq: bool = False,
    **kwargs
) -> DataFrame:
    """
    Executes an SQL query against a Splitgraph image, returning the result.

    Extra `**kwargs` are passed to Pandas' `read_sql_query`.

    :param sql: SQL query to execute.
    :param image: Image object, image hash/tag (`str`) or None (use the currently checked out image).
    :param repository: Repository the image belongs to. Must be set if `image` is a hash/tag or None.
    :param use_lq: Whether to use layered querying or check out the image if it's not checked out.
    :return: A Pandas dataframe.
    """
    return _pandas_adapter.to_data(sql, image, repository, use_lq, **kwargs)


def df_to_table(
    df: Union[Series, DataFrame],
    repository: Repository,
    table: str,
    if_exists: str = "patch",
    schema_check: bool = True,
) -> None:
    """Writes a Pandas DataFrame to a checked-out Splitgraph table. Doesn't create a new image.

    :param df: Pandas DataFrame to insert.
    :param repository: Splitgraph Repository object. Must be checked out.
    :param table: Table name.
    :param if_exists: Behaviour if the table already exists: 'patch' means that primary keys that already exist in the
    table will be updated and ones that don't will be inserted. 'replace' means that the table will be dropped and
    recreated.
    :param schema_check: If False, skips checking that the dataframe is compatible with the target schema.
    """
    _pandas_adapter.to_table(df, repository, table, if_exists, schema_check)



"""
Defines the interface for a Splitgraph engine (a backing database), including running basic SQL commands,
tracking tables for changes and uploading/downloading tables to other remote engines.

By default, Splitgraph is backed by Postgres: see :mod:`splitgraph.engine.postgres` for an example of how to
implement a different engine.
"""
import re
import threading
from abc import ABC
from contextlib import contextmanager
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from psycopg2.sql import SQL, Composed, Identifier

from splitgraph.config import CONFIG
from splitgraph.config.config import ConfigDict, get_all_in_section, get_singleton
from splitgraph.core.types import TableColumn, TableSchema

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine

# List of config flags that are extracted from the global configuration and passed to a given engine
_ENGINE_SPECIFIC_CONFIG = [
    "SG_ENGINE_HOST",
    "SG_ENGINE_PORT",
    "SG_ENGINE_USER",
    "SG_ENGINE_PWD",
    "SG_ENGINE_DB_NAME",
    "SG_ENGINE_POSTGRES_DB_NAME",
    "SG_ENGINE_ADMIN_USER",
    "SG_ENGINE_ADMIN_PWD",
    "SG_ENGINE_FDW_HOST",
    "SG_ENGINE_FDW_PORT",
    "SG_ENGINE_OBJECT_PATH",
    "SG_NAMESPACE",
    "SG_IS_REGISTRY",
    "SG_CHECK_VERSION",
]

# Some engine config keys default to values of other keys if unspecified.
_ENGINE_CONFIG_DEFAULTS = {
    "SG_ENGINE_FDW_HOST": "SG_ENGINE_HOST",
    "SG_ENGINE_FDW_PORT": "SG_ENGINE_PORT",
}


def _prepare_engine_config(config_dict: ConfigDict, name: str = "LOCAL") -> Dict[str, str]:
    result = {}

    # strictly speaking the "config_dict" itself doesn't have the type Dict[str, str]
    # (since it has nested things) but whatever
    subsection: Dict[str, str] = cast(
        Dict[str, str],
        config_dict if name == "LOCAL" else get_all_in_section(config_dict, "remotes")[name],
    )

    for key in _ENGINE_SPECIFIC_CONFIG:
        try:
            result[key] = subsection[key]
        except KeyError:
            try:
                result[key] = subsection[_ENGINE_CONFIG_DEFAULTS[key]]
            except KeyError:
                result[key] = ""
    return result


class ResultShape(Enum):
    """Shape that the result of a query will be coerced to"""

    NONE = 0  # No result expected
    ONE_ONE = 1  # e.g. "row1_val1"
    ONE_MANY = 2  # e.g. ("row1_val1", "row1_val_2")
    MANY_ONE = 3  # e.g. ["row1_val1", "row2_val_1", ...]
    MANY_MANY = 4  # e.g. [("row1_val1", "row1_val_2"), ("row2_val1", "row2_val_2"), ...]


def validate_type(t: str) -> str:
    if not re.match(r"[\w ()\[\]]", t):
        raise ValueError("Invalid type %s!" % t)
    return t


class SavepointStack(threading.local):
    def __init__(self):
        self.stack: List[str] = []


class SQLEngine(ABC):
    """Abstraction for a Splitgraph SQL backend. Requires any overriding classes to implement `run_sql` as well as
    a few other functions. Together with the `information_schema` (part of the SQL standard), this class uses those
    functions to implement some basic database management methods like listing, deleting, creating, dumping
    and loading tables."""

    def __init__(self) -> None:
        self._savepoint_stack = SavepointStack()

    @contextmanager
    def savepoint(self, name: str) -> Iterator[None]:
        """At the beginning of this context manager, a savepoint is initialized and any database
        error that occurs in run_sql results in a rollback to this savepoint rather than the
        rollback of the whole transaction. At exit, the savepoint is released."""
        self.run_sql(SQL("SAVEPOINT ") + Identifier(name))
        self._savepoint_stack.stack.append(name)
        try:
            yield
            # Don't catch any exceptions here: the implementer's run_sql method is supposed
            # to do that and roll back to the savepoint.
        finally:
            # If the savepoint wasn't rolled back to, release it.
            if self._savepoint_stack.stack and self._savepoint_stack.stack[-1] == name:
                self._savepoint_stack.stack.pop()
                self.run_sql(SQL("RELEASE SAVEPOINT ") + Identifier(name))

    def run_sql(self, statement, arguments=None, return_shape=ResultShape.MANY_MANY, named=False):
        """Run an arbitrary SQL statement with some arguments, return an iterator of results.
        If the statement doesn't return any results, return None. If named=True, return named
        tuples when possible."""
        raise NotImplementedError()

    def commit(self):
        """Commit the engine's backing connection"""

    def close(self):
        """Commit and close the engine's backing connection"""

    def rollback(self):
        """Rollback the engine's backing connection"""

    def run_sql_batch(self, statement, arguments, schema=None):
        """Run a parameterized SQL statement against multiple sets of arguments.

        :param statement: Statement to run
        :param arguments: Query arguments
        :param schema: Schema to run the statement in"""
        raise NotImplementedError()

    def run_sql_in(
        self,
        schema: str,
        sql: Union[Composed, str],
        arguments: None = None,
        return_shape: ResultShape = ResultShape.MANY_MANY,
    ) -> Any:
        """
        Executes a non-schema-qualified query against a specific schema.

        :param schema: Schema to run the query in
        :param sql: Query
        :param arguments: Query arguments
        :param return_shape: ReturnShape to coerce the result into.
        """
        self.run_sql("SET search_path TO %s,public", (schema,), return_shape=ResultShape.NONE)
        result = self.run_sql(sql, arguments, return_shape=return_shape)
        self.run_sql("SET search_path TO public", (schema,), return_shape=ResultShape.NONE)
        return result

    def table_exists(self, schema: str, table_name: str) -> bool:
        """
        Check if a table exists on the engine.

        :param schema: Schema name
        :param table_name: Table name
        """
        return (
            self.run_sql(
                """SELECT table_name from information_schema.tables
                           WHERE table_schema = %s AND table_name = %s""",
                (schema, table_name[:63]),
                ResultShape.ONE_ONE,
            )
            is not None
        )

    def schema_exists(self, schema: str) -> bool:
        """
        Check if a schema exists on the engine.

        :param schema: Schema name
        """
        return (
            self.run_sql(
                """SELECT 1 from information_schema.schemata
                           WHERE schema_name = %s""",
                (schema,),
                return_shape=ResultShape.ONE_ONE,
            )
            is not None
        )

    def create_schema(self, schema: str) -> None:
        """Create a schema if it doesn't exist"""
        self.run_sql(SQL("CREATE SCHEMA IF NOT EXISTS {}").format(Identifier(schema)))

    def copy_table(
        self,
        source_schema: str,
        source_table: str,
        target_schema: str,
        target_table: str,
        with_pk_constraints: bool = True,
    ) -> None:
        """Copy a table in the same engine, optionally applying primary key constraints as well."""

        if not self.table_exists(target_schema, target_table):
            query = SQL("CREATE TABLE {}.{} AS SELECT * FROM {}.{}").format(
                Identifier(target_schema),
                Identifier(target_table),
                Identifier(source_schema),
                Identifier(source_table),
            )
        else:
            query = SQL("INSERT INTO {}.{} SELECT * FROM {}.{}").format(
                Identifier(target_schema),
                Identifier(target_table),
                Identifier(source_schema),
                Identifier(source_table),
            )
        pks = self.get_primary_keys(source_schema, source_table)

        if with_pk_constraints and pks:
            query += (
                SQL(";ALTER TABLE {}.{} ADD PRIMARY KEY (").format(
                    Identifier(target_schema), Identifier(target_table)
                )
                + SQL(",").join(SQL("{}").format(Identifier(c)) for c, _ in pks)
                + SQL(")")
            )
        self.run_sql(query)

    def delete_table(self, schema: str, table: str) -> None:
        """Drop a table from a schema if it exists"""
        table_type = self.get_table_type(schema, table)
        # Add CASCADE to drop dependent tables like views etc.
        if table_type == "VIEW":
            self.run_sql(
                SQL("DROP VIEW IF EXISTS {}.{} CASCADE").format(
                    Identifier(schema), Identifier(table)
                )
            )
        elif table_type not in ("FOREIGN TABLE", "FOREIGN"):
            self.run_sql(
                SQL("DROP TABLE IF EXISTS {}.{} CASCADE").format(
                    Identifier(schema), Identifier(table)
                )
            )
        else:
            self.run_sql(
                SQL("DROP FOREIGN TABLE IF EXISTS {}.{}").format(
                    Identifier(schema), Identifier(table)
                )
            )

    def delete_schema(self, schema: str) -> None:
        """Delete a schema if it exists, including all the tables in it."""
        self.run_sql(
            SQL("DROP SCHEMA IF EXISTS {} CASCADE").format(Identifier(schema)),
            return_shape=ResultShape.NONE,
        )

    def get_all_tables(self, schema: str) -> List[str]:
        """Get all tables in a given schema."""
        return cast(
            List[str],
            self.run_sql(
                "SELECT table_name FROM information_schema.tables WHERE table_schema = %s",
                (schema,),
                return_shape=ResultShape.MANY_ONE,
            ),
        )

    def get_table_type(self, schema: str, table: str) -> Optional[str]:
        """Get the type of the table (BASE or FOREIGN)"""
        result = self.run_sql(
            "SELECT table_type FROM information_schema.tables WHERE table_schema = %s"
            " AND table_name = %s",
            (schema, table),
            return_shape=ResultShape.ONE_ONE,
        )
        return cast(Optional[str], result)

    def get_primary_keys(self, schema, table):
        """Get a list of (column_name, column_type) denoting the primary keys of a given table."""
        raise NotImplementedError()

    @staticmethod
    def dump_table_creation(
        schema: Optional[str],
        table: str,
        schema_spec: "TableSchema",
        unlogged: bool = False,
        temporary: bool = False,
        include_comments: bool = False,
    ) -> Tuple[Composed, Tuple]:
        """
        Dumps the DDL for a table using a previously-dumped table schema spec

        :param schema: Schema to create the table in
        :param table: Table name to create
        :param schema_spec: TableSchema
        :param unlogged: If True, the table won't be reflected in the WAL or scanned by the analyzer/autovacuum.
        :param temporary: If True, a temporary table is created (the schema parameter is ignored)
        :param include_comments: If True, also adds COMMENT statements for columns that have them.
        :return: An SQL statement that reconstructs the table schema + args to be mogrified into it.
        """
        flavour = ""
        if unlogged:
            flavour = "UNLOGGED"
        if temporary:
            flavour = "TEMPORARY"

        schema_spec = sorted(schema_spec)

        if temporary:
            target = Identifier(table)
        else:
            target = SQL("{}.{}").format(Identifier(schema), Identifier(table))
        query = (
            SQL("CREATE " + flavour + " TABLE ")
            + target
            + SQL(
                " (" + ",".join("{} %s " % validate_type(col.pg_type) for col in schema_spec)
            ).format(*(Identifier(col.name) for col in schema_spec))
        )

        pk_cols = [col.name for col in schema_spec if col.is_pk]
        if pk_cols:
            query += (
                SQL(", PRIMARY KEY (")
                + SQL(",").join(SQL("{}").format(Identifier(c)) for c in pk_cols)
                + SQL("))")
            )
        else:
            query += SQL(")")
        if unlogged:
            query += SQL(" WITH(autovacuum_enabled=false)")
        query += SQL(";")

        args = []
        if include_comments:
            for col in schema_spec:
                if col.comment:
                    query += SQL("COMMENT ON COLUMN {}.{} IS %s;").format(
                        target, Identifier(col.name)
                    )
                    args.append(col.comment)

        return query, tuple(args)

    def create_table(
        self,
        schema: Optional[str],
        table: str,
        schema_spec: "TableSchema",
        unlogged: bool = False,
        temporary: bool = False,
        include_comments: bool = False,
    ) -> None:
        """
        Creates a table using a previously-dumped table schema spec

        :param schema: Schema to create the table in
        :param table: Table name to create
        :param schema_spec: TableSchema
        :param unlogged: If True, the table won't be reflected in the WAL or scanned by the analyzer/autovacuum.
        :param temporary: If True, a temporary table is created (the schema parameter is ignored)
        :param include_comments: If True, also adds COMMENT statements for columns that have them.
        """
        query, args = self.dump_table_creation(
            schema, table, schema_spec, unlogged, temporary, include_comments
        )
        self.run_sql(query, args)

    def dump_table_sql(
        self,
        schema,
        table_name,
        stream,
        columns="*",
        where="",
        where_args=None,
        target_schema=None,
        target_table=None,
    ):
        """
        Dump the table contents in the SQL format
        :param schema: Schema the table is located in
        :param table_name: Name of the table
        :param stream: A file-like object to write the result into.
        :param columns: SQL column spec. Default '*'.
        :param where: Optional, an SQL WHERE clause
        :param where_args: Arguments for the optional WHERE clause.
        :param target_schema: Schema to create the table in (default same as `schema`)
        :param target_table: Name of the table to insert data into (default same as `table_name`)
        """
        raise NotImplementedError()

    def get_full_table_schema(self, schema: str, table_name: str) -> "TableSchema":
        """
        Generates a list of (column ordinal, name, data type, is_pk, column comment),
        used to detect schema changes like columns being dropped/added/renamed or type changes.

        NB this doesn't work for temporary tables (pg_temp) and returns an empty schema.
        """
        assert schema != "pg_temp"

        results = self.run_sql(
            "SELECT c.attnum, c.attname, "
            "pg_catalog.format_type(c.atttypid, c.atttypmod), pgd.description "
            "FROM pg_attribute c JOIN pg_class t ON c.attrelid = t.oid "
            "JOIN pg_namespace n ON t.relnamespace = n.oid "
            "LEFT JOIN pg_description pgd ON pgd.objoid = t.oid AND pgd.objsubid = c.attnum "
            "WHERE n.nspname = %s AND t.relname = %s AND NOT c.attisdropped "
            "AND c.attnum >= 0 ORDER BY c.attnum ",
            (schema, table_name),
        )

        def _convert_type(ctype):
            # We don't keep a lot of type information, so e.g. char(5) gets turned into char
            # which defaults into char(1).
            return ctype if ctype != "character" else "character varying"

        # Do we need to make sure the PK has the same type + ordinal position here?
        pks = [pk for pk, _ in self.get_primary_keys(schema, table_name)]

        return [TableColumn(o, n, _convert_type(dt), (n in pks), c) for o, n, dt, c in results]

    def initialize(self):
        """Does any required initialization of the engine"""

    def lock_table(self, schema, table):
        """Acquire an exclusive lock on a given table, released when the transaction commits / rolls back."""
        raise NotImplementedError()


class ChangeEngine(SQLEngine, ABC):
    """An SQL engine that can perform change tracking on a set of tables."""

    def get_tracked_tables(self):
        """
        :return: A list of (table_schema, table_name) that the engine currently tracks for changes
        """
        raise NotImplementedError()

    def track_tables(self, tables):
        """
        Start engine-specific change tracking on a list of tables.

        :param tables: List of (table_schema, table_name) to start tracking
        """
        raise NotImplementedError()

    def untrack_tables(self, tables):
        """
        Stop engine-specific change tracking on a list of tables and delete any pending changes.

        :param tables: List of (table_schema, table_name) to start tracking
        """
        raise NotImplementedError()

    def has_pending_changes(self, schema):
        """
        Return True if the tracked schema has pending changes and False if it doesn't.
        """
        raise NotImplementedError()

    def discard_pending_changes(self, schema, table=None):
        """
        Discard recorded pending changes for a tracked table or the whole schema
        """
        raise NotImplementedError()

    def get_pending_changes(self, schema, table, aggregate=False):
        """
        Return pending changes for a given tracked table

        :param schema: Schema the table belongs to
        :param table: Table to return changes for
        :param aggregate: Whether to aggregate changes or return them completely
        :return: If aggregate is True: tuple with numbers of `(added_rows, removed_rows, updated_rows)`.
            If aggregate is False: A changeset. The changeset is a list of
            `(pk, action (0 for Insert, 1 for Delete, 2 for Update), action_data)`
            where `action_data` is `None` for Delete and `{'c': [column_names], 'v': [column_values]}` that
            have been inserted/updated otherwise.
        """
        raise NotImplementedError()

    def get_changed_tables(self, schema):
        """
        List tracked tables that have pending changes

        :param schema: Schema to check for changes
        :return: List of tables with changed contents
        """
        raise NotImplementedError()

    def get_change_key(self, schema: str, table: str) -> List[Tuple[str, str]]:
        """
        Returns the key used to identify a row in a change (list of column name, column type).
        If the tracked table has a PK, we use that; if it doesn't, the whole row is used.
        """
        raise NotImplementedError()


class ObjectEngine:
    """
    Routines for storing/applying objects as well as sharing them with other engines.
    """

    def get_object_schema(self, object_id):
        """
        Get the schema of a given object, returned as a list of
        (ordinal, column_name, column_type, is_pk).

        :param object_id: ID of the object
        """

    def get_object_size(self, object_id):
        """
        Return the on-disk footprint of this object, in bytes
        :param object_id: ID of the object
        """

    def delete_objects(self, object_ids):
        """
        Delete one or more objects from the engine.

        :param object_ids: IDs of objects to delete
        """

    def store_fragment(
        self, inserted, deleted, schema, table, source_schema, source_table, source_schema_spec
    ):
        """
        Store a fragment of a changed table in another table

        :param inserted: List of PKs that have been updated/inserted
        :param deleted: List of PKs that have been deleted
        :param schema: Schema to store the change in
        :param table: Table to store the change in
        :param source_schema: Schema the source table is located in
        :param source_table: Name of the source table
        :param source_schema_spec: Schema of the source table (optional)
        """
        raise NotImplementedError()

    def apply_fragments(
        self,
        objects,
        target_schema,
        target_table,
        extra_quals=None,
        extra_qual_args=None,
        schema_spec=None,
        progress_every: Optional[int] = None,
    ):
        """
        Apply multiple fragments to a target table as a single-query batch operation.

        :param objects: List of tuples `(object_schema, object_table)` that the objects are stored in.
        :param target_schema: Schema to apply the fragment to
        :param target_table: Table to apply the fragment to
        :param extra_quals: Optional, extra SQL (Composable) clauses to filter new rows in the fragment on
            (e.g. SQL("a = %s"))
        :param extra_qual_args: Optional, a tuple of arguments to use with `extra_quals`
        :param schema_spec: Optional, list of (ordinal, column_name, column_type, is_pk).
            If not specified, uses the schema of target_table.
        :param progress_every: If set, will report the materialization progress via
            tqdm every `progress_every` objects.
        """
        raise NotImplementedError()

    def upload_objects(self, objects, remote_engine):
        """
        Upload objects from the local cache to the remote engine

        :param objects: List of object IDs to upload
        :param remote_engine: A remote ObjectEngine to upload the objects to.
        """
        raise NotImplementedError()

    def download_objects(self, objects, remote_engine):
        """
        Download objects from the remote engine to the local cache

        :param objects: List of object IDs to download
        :param remote_engine: A remote ObjectEngine to download the objects from.

        :return List of object IDs that were downloaded.
        """
        raise NotImplementedError()

    def dump_object(self, object_id, stream, schema):
        """
        Dump an object into a series of SQL statements

        :param object_id: Object ID
        :param stream: Text stream to dump the object into
        :param schema: Schema the object lives in
        """
        raise NotImplementedError()

    def store_object(
        self,
        object_id: str,
        source_query: Union[bytes, Composed, str, SQL],
        schema_spec: TableSchema,
        source_query_args: Optional[Sequence[Any]],
        overwrite: bool,
    ):
        """
        Stores a Splitgraph object using a source query in the actual format
        implemented by this engine.

        :param object_id: Name of the object
        :param source_query: SELECT query that produces data required by the object
        :param schema_spec: Schema of the source table
        :param source_query_args: Arguments to mogrify into the source query.
        :param overwrite: If True, will overwrite the object if it already exists.
        """


# Name of the current global engine, 'LOCAL' for the local.
# Can be overridden via normal configuration routes, e.g.
# $ SG_ENGINE=remote_engine sgr init
# will initialize the remote engine instead.
_ENGINE: Union[str, "PostgresEngine"] = get_singleton(CONFIG, "SG_ENGINE") or "LOCAL"

# Map of engine names -> Engine instances
_ENGINES: Dict[str, "PostgresEngine"] = {}


def get_engine(
    name: Optional[str] = None,
    use_socket: bool = False,
    use_fdw_params: bool = False,
    autocommit: bool = False,
) -> "PostgresEngine":
    """
    Get the current global engine or a named remote engine

    :param name: Name of the remote engine as specified in the config. If None, the current global engine
        is returned.
    :param use_socket: Use a local UNIX socket instead of PG_HOST, PG_PORT for LOCAL engine connections.
    :param use_fdw_params: Use the _FDW connection parameters (SG_ENGINE_FDW_HOST/PORT). By default,
        will infer from the global splitgraph.config.IN_FDW flag.
    :param autocommit: If True, the engine will not open SQL transactions implicitly.
    """
    from .postgres.engine import PostgresEngine

    if not name:
        if isinstance(_ENGINE, PostgresEngine):
            return _ENGINE
        name = _ENGINE
    if name not in _ENGINES:
        # Here we'd get the engine type/backend (Postgres/MySQL etc)
        # and instantiate the actual Engine class.
        # As we only have PostgresEngine, we instantiate that.

        conn_params = cast(Dict[str, Optional[str]], _prepare_engine_config(CONFIG, name))

        try:
            is_registry = bool(conn_params.pop("SG_IS_REGISTRY"))
        except KeyError:
            is_registry = False
        try:
            check_version = bool(conn_params.pop("SG_CHECK_VERSION"))
        except KeyError:
            check_version = False

        if name == "LOCAL" and use_socket:
            conn_params["SG_ENGINE_HOST"] = None
            conn_params["SG_ENGINE_PORT"] = None
        if use_fdw_params:
            conn_params["SG_ENGINE_HOST"] = conn_params["SG_ENGINE_FDW_HOST"]
            conn_params["SG_ENGINE_PORT"] = conn_params["SG_ENGINE_FDW_PORT"]

        _ENGINES[name] = PostgresEngine(
            conn_params=conn_params,
            name=name,
            autocommit=autocommit,
            registry=is_registry,
            check_version=check_version,
            in_fdw=use_fdw_params and name == "LOCAL",
        )
    return _ENGINES[name]


def set_engine(engine: "PostgresEngine") -> None:
    """
    Switch the global engine to a different one.

    :param engine: Engine
    """
    global _ENGINE
    _ENGINE = engine


@contextmanager
def switch_engine(engine: "PostgresEngine") -> Iterator[None]:
    """
    Switch the global engine to a different one. The engine will
    get switched back on exit from the context manager.

    :param engine: Engine
    """
    global _ENGINE
    _prev_engine = _ENGINE
    try:
        _ENGINE = engine
        yield
    finally:
        _ENGINE = _prev_engine



import json
import logging
import os
import re
from abc import ABC
from contextlib import contextmanager
from random import getrandbits
from typing import Dict, Generator, List, Optional, Tuple

import docker.errors
import pydantic
from docker import DockerClient
from docker.models.containers import Container
from docker.types import LogConfig

from splitgraph.core.repository import Repository
from splitgraph.core.types import (
    IntrospectionResult,
    SyncState,
    TableInfo,
    TableParams,
    get_table_params,
)
from splitgraph.engine.postgres.engine import PostgresEngine
from splitgraph.hooks.data_source.base import (
    SyncableDataSource,
    get_ingestion_state,
    prepare_new_image,
)
from splitgraph.utils.docker import copy_to_container, get_docker_client

from ..common import add_timestamp_tags
from ..dbt.utils import run_dbt_transformation_from_git
from ..singer.common import store_ingestion_state
from .docker_utils import (
    add_files,
    build_command,
    detect_network_mode,
    remove_at_end,
    wait_not_failed,
)
from .models import AirbyteCatalog, AirbyteMessage, ConfiguredAirbyteCatalog
from .utils import (
    AirbyteConfig,
    _airbyte_message_reader,
    _store_processed_airbyte_tables,
    _store_raw_airbyte_tables,
    get_pk_cursor_fields,
    get_sg_schema,
    select_streams,
)


@contextmanager
def delete_schema_at_end(engine: "PostgresEngine", schema: str) -> Generator:
    try:
        yield
    finally:
        engine.delete_schema(schema)
        engine.commit()


class AirbyteDataSource(SyncableDataSource, ABC):
    """Generic data source for Airbyte-compliant sources.
    We run ingestion by combining an Airbyte source and the Airbyte Postgres destination.
    """

    docker_image: Optional[str] = None
    airbyte_name: Optional[str] = None
    receiver_image = "airbyte/destination-postgres:0.3.12"
    normalization_image = "airbyte/normalization:0.1.59"
    cursor_overrides: Optional[Dict[str, List[str]]] = None
    primary_key_overrides: Optional[Dict[str, List[str]]] = None
    docker_environment: Optional[Dict[str, str]] = None

    table_params_schema = {
        "type": "object",
        "properties": {
            "airbyte_cursor_field": {
                "type": "array",
                "title": "Cursor field(s)",
                "description": "Fields in this stream to be used as a cursor "
                "for incremental replication (overrides Airbyte configuration's cursor_field)",
                "items": {"type": "string"},
            },
            "airbyte_primary_key_field": {
                "type": "array",
                "title": "Primary key field(s)",
                "description": "Fields in this stream to be used as a primary key for deduplication "
                "(overrides Airbyte configuration's primary_key)",
                "items": {"type": "string"},
            },
        },
    }

    # Base parameters/credentials for this data source (allowing users to pass in a Git repo
    # with a dbt project). Note that classes inheriting this should keep this JSONSchema by
    # redefining it as:
    #
    #   credentials_schema = merge_jsonschema(
    #       AirbyteDataSource.credentials_schema,
    #       {"type": "object", "properties": {...}},
    #   )
    #   params_schema = merge_jsonschema(
    #       AirbyteDataSource.params_schema,
    #       {"type": "object", "properties": {...}},
    #   )
    params_schema = {
        "type": "object",
        "properties": {
            "normalization_mode": {
                "type": "string",
                "title": "Post-ingestion normalization",
                "description": "Whether to normalize raw Airbyte tables. "
                "`none` is no normalization, `basic` is Airbyte's basic normalization, "
                "`custom` is a custom dbt transformation on the data.",
                "enum": ["none", "basic", "custom"],
                "default": "basic",
            },
            "normalization_git_branch": {
                "type": "string",
                "title": "dbt model Git branch",
                "description": "Branch or commit hash to use for the normalization dbt project.",
                "default": "master",
            },
        },
    }

    credentials_schema = {
        "type": "object",
        "properties": {
            # This is a secret since the git URL might have a password in it.
            "normalization_git_url": {
                "type": "string",
                "title": "dbt model Git URL",
                "description": "For `custom` normalization, a URL to the Git repo "
                "with the dbt project, for example,"
                "`https://uname:pass_or_token@github.com/organisation/repository.git`.",
            },
        },
    }

    _internal_params = ["normalization_mode", "normalization_git_branch", "normalization_git_url"]

    def get_airbyte_config(self) -> AirbyteConfig:
        return {
            k: v
            for k, v in {**self.params, **self.credentials}.items()
            if k not in self._internal_params
        }

    def _sync(
        self,
        schema: str,
        state: Optional[SyncState] = None,
        tables: Optional[TableInfo] = None,
    ) -> SyncState:
        # We override the main sync() instead
        pass

    def load(self, repository: "Repository", tables: Optional[TableInfo] = None) -> str:
        return self.sync(repository, image_hash=None, tables=tables, use_state=False)

    @staticmethod
    def _make_postgres_config(engine: PostgresEngine, schema: str) -> AirbyteConfig:
        return {
            "host": engine.conn_params["SG_ENGINE_HOST"],
            "port": int(engine.conn_params["SG_ENGINE_PORT"] or 5432),
            "username": engine.conn_params["SG_ENGINE_USER"],
            "password": engine.conn_params["SG_ENGINE_PWD"],
            "database": engine.conn_params["SG_ENGINE_DB_NAME"],
            "schema": schema,
            "ssl": False,
        }

    def _run_discovery(self, config: Optional[AirbyteConfig] = None) -> AirbyteCatalog:
        client = get_docker_client()
        network_mode = detect_network_mode()

        with self._source_container(
            client,
            network_mode=network_mode,
            config=config,
            catalog=None,
            state=None,
            discover=True,
        ) as container:
            # Copy config into /
            copy_to_container(
                container,
                source_path=None,
                target_path="/config.json",
                data=json.dumps(config or {}).encode(),
            )

            container.start()
            wait_not_failed(container, mirror_logs=False)

            # Grab the catalog from the output (it's mixed with other logs)
            for message in _airbyte_message_reader(container.logs(stream=True)):
                if message.catalog:
                    logging.info("Catalog: %s", message.catalog)
                    return message.catalog
        raise AssertionError("No catalog output!")

    def sync(
        self,
        repository: Repository,
        image_hash: Optional[str] = None,
        tables: Optional[TableInfo] = None,
        use_state: bool = True,
    ) -> str:
        # https://docs.airbyte.io/understanding-airbyte/airbyte-specification
        self._validate_table_params(tables)
        tables = tables or self.tables

        # Select columns and streams (full_refresh/incremental, cursors)
        src_config = self.get_airbyte_config()
        catalog = self._run_discovery(src_config)
        configured_catalog = select_streams(
            catalog,
            tables,
            sync=use_state,
            cursor_overrides=self.cursor_overrides,
            primary_key_overrides=self.primary_key_overrides,
        )
        logging.info("Configured catalog: %s", configured_catalog)

        # Create a destination catalog that overrides the namespace in the source to None.
        # Some sources and the PG destination respect stream.namespace -- in the case of the
        # MySQL source, it denotes the source database name and for PG, it's the target
        # schema name. We need to let the source keep its old namespace and override the
        # destination for PG (set it to None here and inject it into the config).
        dst_catalog = configured_catalog.copy(deep=True)
        for stream in dst_catalog.streams:
            stream.stream.namespace = None

        # Load ingestion state
        base_image, new_image_hash = prepare_new_image(
            repository, image_hash, comment="Airbyte data load"
        )
        state = get_ingestion_state(repository, image_hash) if use_state else None
        logging.info("Current ingestion state: %s", state)

        # Set up a staging schema for the data
        # Delete the slashes or Airbyte will do it for us.
        staging_schema = "sg_tmp_" + repository.to_schema().replace("/", "_").replace("-", "_")
        dst_config = self._make_postgres_config(repository.object_engine, staging_schema)
        client = get_docker_client()
        network_mode = detect_network_mode()

        with delete_schema_at_end(repository.object_engine, staging_schema):
            repository.object_engine.delete_schema(staging_schema)
            repository.object_engine.create_schema(staging_schema)
            repository.commit_engines()
            # Run the Airbyte source and receiver and pipe data between them, writing it
            # out into a temporary schema.

            logging.info("Running Airbyte EL process")
            dest_files, new_state, sync_modes = self._run_airbyte_el(
                client, network_mode, src_config, dst_config, configured_catalog, dst_catalog, state
            )

            # At this stage, Airbyte wrote out the raw tables into the staging schema: they have
            # the form _airbyte_tmp_STREAM_NAME and schema (hash, raw_json, date). These raw tables
            # are append-or-truncate only, so we append/replace them in the existing Splitgraph image
            # at this stage.

            logging.info("Storing raw tables as Splitgraph images")
            raw_tables = _store_raw_airbyte_tables(
                repository,
                new_image_hash,
                staging_schema,
                sync_modes,
                default_sync_mode="append_dedup" if use_state else "overwrite",
            )

            # We can also store the state at this point, since only the raw tables depend on it.
            store_ingestion_state(
                repository,
                new_image_hash,
                current_state=state,
                new_state=json.dumps(new_state) if new_state else "{}",
            )
            add_timestamp_tags(repository, new_image_hash)
            repository.commit_engines()

        normalization_mode = self.params.get("normalization_mode", "basic")

        if normalization_mode == "none":
            return new_image_hash

        # Run normalization
        # This converts the raw Airbyte tables (with JSON) into actual tables with fields.
        # We first replace the raw table fragments that Airbyte wrote out with the actual full
        # tables, checked out via LQ so that dbt (run by Airbyte's normalization container) can
        # scan through them and build the actual ingested data.

        new_image = repository.images.by_hash(new_image_hash)
        with delete_schema_at_end(repository.object_engine, staging_schema):
            repository.object_engine.create_schema(staging_schema)
            new_image.lq_checkout(staging_schema, only_tables=raw_tables)
            repository.commit_engines()
            logging.info("Running Airbyte T step (normalization)")

            if normalization_mode == "basic":
                # Run Airbyte's basic normalization container that autogenerates a dbt model from
                # the catalog file.
                # This actually always recreates the normalized tables from scratch.
                # https://github.com/airbytehq/airbyte/issues/4286
                logging.info("Using basic normalization")
                with self._normalization_container(client, network_mode) as normalization_container:
                    add_files(normalization_container, dest_files)
                    normalization_container.start()
                    wait_not_failed(normalization_container, mirror_logs=True)
            else:
                logging.info("Using a dbt project from Git")

                try:
                    git_url = self.credentials["normalization_git_url"]
                except KeyError:
                    raise ValueError("No normalization_git_url specified in plugin credentials!")
                git_ref = self.params.get("normalization_git_branch", "master")

                run_dbt_transformation_from_git(
                    repository.object_engine, staging_schema, git_url, git_ref
                )

            logging.info("Storing normalized tables")
            _store_processed_airbyte_tables(repository, new_image_hash, staging_schema)
            repository.commit_engines()

        return new_image_hash

    def _run_airbyte_el(
        self,
        client: docker.DockerClient,
        network_mode: str,
        src_config: AirbyteConfig,
        dst_config: AirbyteConfig,
        src_catalog: ConfiguredAirbyteCatalog,
        dst_catalog: ConfiguredAirbyteCatalog,
        state: Optional[SyncState],
    ) -> Tuple[List[Tuple[str, str]], Optional[SyncState], Dict[str, str]]:
        with self._source_container(
            client, network_mode, src_config, src_catalog, state
        ) as source, self._destination_container(
            client, network_mode, dst_config, src_catalog
        ) as destination:

            # Set up the files in src/dest containers
            add_files(
                source,
                [
                    ("config", json.dumps(src_config)),
                    (
                        "catalog",
                        src_catalog.json(exclude_unset=True, exclude_defaults=True),
                    ),
                    ("state", json.dumps(state)),
                ],
            )

            dest_files = [
                ("config", json.dumps(dst_config)),
                (
                    "catalog",
                    dst_catalog.json(exclude_unset=True, exclude_defaults=True),
                ),
            ]

            add_files(destination, dest_files)

            dest_socket = destination.attach_socket(params={"stdin": 1, "stream": 1})
            dest_socket._writing = True
            src_socket = source.attach(stdout=True, stream=True, logs=True)

            source.start()
            destination.start()

            # Pipe messages from the source to the destination
            for message in _airbyte_message_reader(src_socket):
                if message.state or message.record:
                    if message.record:
                        # Empty out the namespace so that we use the destination schema in PG
                        message.record.namespace = None

                    out = (message.json(exclude_unset=True, exclude_defaults=True) + "\n").encode()
                    logging.debug("Writing message %s", out)
                    while out:
                        written = dest_socket.write(out)
                        out = out[written:]

                    dest_socket.flush()
                elif message.log:
                    logging.info(message.log.message)

            # NB this is the magic thing that makes the socket actually close and kick the container so that
            # it sees that STDIN is closed too.
            # Neither of these work
            #   dest_socket.close()
            #   dest_socket._sock.close()
            # Thank you Docker.
            # https://github.com/d11wtq/dockerpty/blob/f8d17d893c6758b7cc25825e99f6b02202632a97/dockerpty/io.py#L182
            # https://github.com/docker/docker-py/issues/1507
            # https://github.com/docker/docker-py/issues/983#issuecomment-492513718
            os.close(dest_socket._sock.fileno())

            wait_not_failed(source)
            wait_not_failed(destination)
            dest_logs = destination.logs(stream=True)

            # Grab the state from stdout
            new_state: Optional[SyncState] = None
            table_sync_modes: Dict[str, str] = {}

            for line in dest_logs:
                line = line.decode()
                logging.info("%s: %s", destination.name, line)

                # Another thing we want to find out from the destination is how it normalized
                # raw stream names (which can be any UTF-8 string) into the output table names
                # (_airbyte_raw_xxx) and the sync mode (overwrite/append). This is because
                # we get Airbyte to always write into empty tables (merging them into the full
                # Splitgraph tables after the fact) but we need to know if it meant to truncate
                # or append to those tables.
                # The PG destination outputs a log message in this format:
                #
                # Write config: WriteConfig{streamName=sites, namespace=null, outputSchemaName=sg_tmp_airbyte_google_test, tmpTableName=_airbyte_tmp_cav_sites, outputTableName=_airbyte_raw_sites, syncMode=overwrite}
                #
                # So we can grab the outputTableName and syncMode to find these out.
                #
                # Other ways of doing this: detect TRUNCATE on our tables (this is probably the best
                # long-term solution, since we want to turn this into just writing to the DDN);
                # back out the table names from the stream names by reimplementing/copying
                # https://github.com/airbytehq/airbyte/blob/441435a373f03262ce87a53505b1863d5554cc6c/airbyte-integrations/bases/base-normalization/normalization/transform_catalog/destination_name_transformer.py#L53.
                match = re.match(r".*outputTableName=([^,]+), syncMode=(\w+)", line)
                if match:
                    raw_table, sync_mode = match.groups()
                    table_sync_modes[raw_table] = sync_mode

                # Also find the STATE message in the log denoting the new connector bookmark.
                if not line.startswith("{"):
                    continue
                try:
                    message = AirbyteMessage.parse_raw(line)
                except pydantic.ValidationError:
                    logging.warning("Couldn't parse message, continuing")
                    continue
                if message.state:
                    new_state = SyncState(message.state.data)
        logging.info("New state: %s", new_state)
        return dest_files, new_state, table_sync_modes

    @contextmanager
    def _source_container(
        self,
        client: DockerClient,
        network_mode: str,
        config: Optional[AirbyteConfig],
        catalog: Optional[ConfiguredAirbyteCatalog],
        state: Optional[SyncState],
        discover: bool = False,
    ) -> Container:
        client.images.pull(self.docker_image)
        container_name = "sg-ab-src-{:08x}".format(getrandbits(64))
        if discover:
            command = ["discover"] + build_command([("config", config)])
        else:
            command = ["read"] + build_command(
                [("config", config), ("state", state), ("catalog", catalog)]
            )
        container = client.containers.create(
            image=self.docker_image,
            name=container_name,
            command=command,
            network_mode=network_mode,
            environment=self.docker_environment,
            log_config=LogConfig(type="json-file", config={"max-size": "10m", "max-file": "3"}),
        )
        with remove_at_end(container):
            yield container

    @contextmanager
    def _destination_container(
        self,
        client: DockerClient,
        network_mode: str,
        config: AirbyteConfig,
        catalog: ConfiguredAirbyteCatalog,
    ) -> Container:
        # Create the Postgres receiver container
        client.images.pull(self.receiver_image)
        destination_container_name = "sg-ab-dst-{:08x}".format(getrandbits(64))
        command = ["write"] + build_command([("config", config), ("catalog", catalog)])
        container = client.containers.create(
            image=self.receiver_image,
            name=destination_container_name,
            command=command,
            network_mode=network_mode,
            stdin_open=True,
            environment=self.docker_environment,
            log_config=LogConfig(type="json-file", config={"max-size": "10m", "max-file": "3"}),
        )
        with remove_at_end(container):
            yield container

    @contextmanager
    def _normalization_container(self, client: DockerClient, network_mode: str) -> Container:
        client.images.pull(self.normalization_image)
        # https://github.com/airbytehq/airbyte/blob/830fac6b648263e1add3589294fcabf4bee6fd39/airbyte-workers/src/main/java/io/airbyte/workers/normalization/DefaultNormalizationRunner.java#L111
        command = [
            "run",
            "--integration-type",
            "postgres",
            "--config",
            "/config.json",
            "--catalog",
            "/catalog.json",
        ]
        container = client.containers.create(
            image=self.normalization_image,
            name="sg-ab-norm-{:08x}".format(getrandbits(64)),
            command=command,
            network_mode=network_mode,
            environment=self.docker_environment,
            log_config=LogConfig(type="json-file", config={"max-size": "10m", "max-file": "3"}),
        )

        with remove_at_end(container):
            yield container

    def introspect(self) -> IntrospectionResult:
        config = self.get_airbyte_config()
        catalog = self._run_discovery(config)

        result = IntrospectionResult({})
        for stream in catalog.streams:
            stream_name = stream.name
            stream_schema = get_sg_schema(stream)

            cursor_field, primary_key = get_pk_cursor_fields(
                stream,
                get_table_params(self.tables, stream.name) if self.tables else TableParams({}),
                self.cursor_overrides,
                self.primary_key_overrides,
            )

            suggested_params = {}
            if cursor_field is not None:
                suggested_params["airbyte_cursor_field"] = cursor_field
            if primary_key is not None:
                suggested_params["airbyte_primary_key"] = [k[0] for k in primary_key]

            result[stream_name] = (stream_schema, TableParams(suggested_params))
        return result



"""Splitgraph mount handler for Socrata datasets"""
import json
import logging
from copy import deepcopy
from typing import Dict, List, Optional, Tuple

from psycopg2.sql import SQL, Identifier

from splitgraph.core.types import Credentials, MountError, TableInfo
from splitgraph.exceptions import RepositoryNotFoundError
from splitgraph.hooks.data_source.fdw import (
    ForeignDataWrapperDataSource,
    create_foreign_table,
)


class SocrataDataSource(ForeignDataWrapperDataSource):
    credentials_schema = {
        "type": "object",
        "properties": {"app_token": {"type": "string", "description": "Socrata app token"}},
    }

    params_schema = {
        "type": "object",
        "properties": {
            "domain": {
                "type": "string",
                "description": "Socrata domain, for example, data.albanyny.gov",
            },
            "batch_size": {
                "type": "integer",
                "description": "Amount of rows to fetch from Socrata per request (limit parameter)",
                "minimum": 1,
                "default": 1000,
                "maximum": 50000,
            },
        },
        "required": ["domain"],
    }

    table_params_schema = {
        "type": "object",
        "properties": {
            "socrata_id": {"type": "string", "description": "Socrata dataset ID, e.g. xzkq-xp2w"}
        },
        "required": ["socrata_id"],
    }

    @classmethod
    def get_name(cls) -> str:
        return "Socrata"

    @classmethod
    def get_description(cls) -> str:
        return "Data source for remote Socrata datasets that uses SoQL for live queries"

    def get_fdw_name(self):
        return "multicorn"

    @classmethod
    def from_commandline(cls, engine, commandline_kwargs) -> "SocrataDataSource":
        params = deepcopy(commandline_kwargs)

        # Convert the old-style "tables" param ({"table_name": "some_id"})
        # to the schema this data source expects
        #   {"table_name": [table schema], {"socrata_id": "some_id"}}
        # Note that we don't actually care about the table schema in this data source,
        # since it reintrospects on every mount.

        tables = params.pop("tables", [])
        if isinstance(tables, dict) and isinstance(next(iter(tables.values())), str):
            tables = {k: ([], {"socrata_id": v}) for k, v in tables.items()}

        credentials = Credentials({})
        return cls(engine, credentials, params, tables)

    def get_server_options(self):
        options: Dict[str, Optional[str]] = {
            "wrapper": "splitgraph.ingestion.socrata.fdw.SocrataForeignDataWrapper"
        }
        for k in ["domain", "batch_size"]:
            if self.params.get(k):
                options[k] = str(self.params[k])
        if self.credentials.get("app_token"):
            options["app_token"] = str(self.credentials["app_token"])
        return options

    def _create_foreign_tables(
        self, schema: str, server_id: str, tables: TableInfo
    ) -> List[MountError]:
        from psycopg2.sql import SQL
        from sodapy import Socrata

        logging.info("Getting Socrata metadata")
        client = Socrata(domain=self.params["domain"], app_token=self.credentials.get("app_token"))

        tables = self.tables or tables
        if isinstance(tables, list):
            sought_ids = tables
        else:
            sought_ids = [t[1]["socrata_id"] for t in tables.values()]

        try:
            datasets = client.datasets(ids=sought_ids, only=["dataset"])
        except Exception as e:
            if "Unknown response format: text/html" in str(e):
                # If the Socrata dataset/domain isn't found, sodapy doesn't catch it directly
                # and instead stumbles on an unexpected content-type of the 404 page it's served.
                # We catch that and reraise a more friendly message.
                raise RepositoryNotFoundError("Socrata domain or dataset not found!") from e
            raise

        if not datasets:
            raise RepositoryNotFoundError("Socrata domain or dataset not found!")

        mount_statements, mount_args = generate_socrata_mount_queries(
            sought_ids, datasets, schema, server_id, tables
        )

        self.engine.run_sql(SQL(";").join(mount_statements), mount_args)
        return []

    def get_raw_url(
        self, tables: Optional[TableInfo] = None, expiry: int = 3600
    ) -> Dict[str, List[Tuple[str, str]]]:
        tables = tables or self.tables
        if not tables:
            return {}
        result: Dict[str, List[Tuple[str, str]]] = {}

        for table in tables:
            table_options = super().get_table_options(table, tables)
            full_options = {**self.params, **table_options}
            domain = full_options.get("domain")
            socrata_id = full_options.get("socrata_id")
            if not domain or not socrata_id:
                continue

            result[table] = [
                (
                    "text/csv",
                    f"https://{domain}/api/views/{socrata_id}/rows.csv?accessType=DOWNLOAD",
                )
            ]

        return result


def generate_socrata_mount_queries(sought_ids, datasets, mountpoint, server_id, tables: TableInfo):
    # Local imports since this module gets run from commandline entrypoint on startup.

    from splitgraph.core.output import pluralise, slugify
    from splitgraph.ingestion.socrata.querying import socrata_to_sg_schema

    found_ids = {d["resource"]["id"] for d in datasets}
    logging.info("Loaded metadata for %s", pluralise("Socrata table", len(found_ids)))

    tables_inv = _get_table_map(found_ids, sought_ids, tables)

    mount_statements = []
    mount_args = []
    for dataset in datasets:
        socrata_id = dataset["resource"]["id"]
        table_name = tables_inv.get(socrata_id) or slugify(
            dataset["resource"]["name"]
        ) + "_" + socrata_id.replace("-", "_")
        schema_spec, column_map = socrata_to_sg_schema(dataset)
        sql, args = create_foreign_table(
            schema=mountpoint,
            server=server_id,
            table_name=table_name,
            schema_spec=schema_spec,
            extra_options={"column_map": json.dumps(column_map), "table": socrata_id},
        )

        description = dataset["resource"].get("description")
        if description:
            sql += SQL("COMMENT ON FOREIGN TABLE {}.{} IS %s").format(
                Identifier(mountpoint), Identifier(table_name)
            )
            args.append(description)

        mount_statements.append(sql)
        mount_args.extend(args)

    return mount_statements, mount_args


def _get_table_map(found_ids, sought_ids, tables: TableInfo) -> Dict[str, str]:
    """Get a map of Socrata ID -> local table name"""
    from splitgraph.core.output import truncate_list

    if isinstance(tables, (dict, list)) and tables:
        missing_ids = [d for d in found_ids if d not in sought_ids]
        if missing_ids:
            raise ValueError(
                "Some Socrata tables couldn't be found! Missing tables: %s"
                % truncate_list(missing_ids)
            )

        if isinstance(tables, dict):
            return {to["socrata_id"]: p for p, (ts, to) in tables.items()}
    return {}



BULK_UPSERT_REPO_PROFILES = """mutation BulkUpsertRepoProfilesMutation(
  $namespaces: [String!]
  $repositories: [String!]
  $descriptions: [String]
  $readmes: [String]
  $licenses: [String]
  $metadata: [JSON]
) {
  __typename
  bulkUpsertRepoProfiles(
  input: {
      namespaces: $namespaces
      repositories: $repositories
      descriptions: $descriptions
      readmes: $readmes
      licenses: $licenses
      metadata: $metadata
  }
  ) {
    clientMutationId
    __typename
  }
}
"""

BULK_UPDATE_REPO_SOURCES = """mutation BulkUpdateRepoSourcesMutation(
  $namespaces: [String!]
  $repositories: [String!]
  $sources: [DatasetSourceInput]
) {
  __typename
  bulkUpdateRepoSources(
  input: {
      namespaces: $namespaces
      repositories: $repositories
      sources: $sources
  }
  ) {
    clientMutationId
    __typename
  }
}
"""

BULK_UPSERT_REPO_TOPICS = """mutation BulkUpsertRepoTopicsMutation(
  $namespaces: [String!]
  $repositories: [String!]
  $topics: [String]
) {
  __typename
  bulkUpsertRepoTopics(
  input: {
      namespaces: $namespaces
      repositories: $repositories
      topics: $topics
  }
  ) {
    clientMutationId
    __typename
  }
}
"""

PROFILE_UPSERT = """mutation UpsertRepoProfile(
  $namespace: String!
  $repository: String!
  $description: String
  $readme: String
  $topics: [String]
  $sources: [DatasetSourceInput]
  $license: String
  $metadata: JSON
) {
  __typename
  upsertRepoProfileByNamespaceAndRepository(
    input: {
      repoProfile: {
        namespace: $namespace
        repository: $repository
        description: $description
        readme: $readme
        sources: $sources
        license: $license
        metadata: $metadata
      }
      patch: {
        namespace: $namespace
        repository: $repository
        description: $description
        readme: $readme
        sources: $sources
        license: $license
        metadata: $metadata
      }
    }
  ) {
    clientMutationId
    __typename
  }
  __typename
  createRepoTopicsAgg(
  input: {
    repoTopicsAgg: {
      namespace: $namespace
      repository: $repository
      topics: $topics
    }
  }
  ) {
    clientMutationId
    __typename
  }
}
"""

FIND_REPO = """query FindRepositories($query: String!, $limit: Int!) {
  findRepository(query: $query, first: $limit) {
    edges {
      node {
        namespace
        repository
        highlight
      }
    }
    totalCount
  }
}"""

REPO_PARAMS = "($namespace: String!, $repository: String!)"
REPO_CONDITIONS = ", condition: {namespace: $namespace, repository: $repository}"
GET_REPO_METADATA = """query GetRepositoryMetadata%s {
  repositories(first: 1000%s) {
    nodes {
      namespace
      repository
      repoTopicsByNamespaceAndRepository {
        nodes {
          topic
        }
      }
      repoProfileByNamespaceAndRepository {
        description
        license
        metadata
        readme
        sources {
          anchor
          href
          isCreator
          isSameAs
        }
      }
    }
  }
}"""

GET_REPO_SOURCE = """query GetRepositoryDataSource%s {
  repositoryDataSources(first: 1000%s) {
    nodes {
      namespace
      repository
      credentialId
      dataSource
      params
      tableParams
      externalImageByNamespaceAndRepository {
        imageByNamespaceAndRepositoryAndImageHash {
          tablesByNamespaceAndRepositoryAndImageHash {
            nodes {
              tableName
              tableSchema
            }
          }
        }
      }
      ingestionScheduleByNamespaceAndRepository {
        schema
        schedule
        enabled
      }
    }
  }
}"""

INGESTION_JOB_STATUS = """
query RepositoryIngestionJobStatus($namespace: String, $repository: String) {
  repositoryIngestionJobStatus(
    first: 1
    namespace: $namespace
    repository: $repository
  ) {
    nodes {
      taskId
      started
      finished
      isManual
      status
    }
  }
}
"""

JOB_LOGS = """
query JobLogs($namespace: String!, $repository: String!, $taskId: String!) {
  jobLogs(namespace: $namespace, repository: $repository, taskId: $taskId) {
    url
  }
}
"""

CSV_URL = """query CSVURLs {
  csvUploadDownloadUrls {
    upload
    download
  }
}
"""

START_LOAD = """mutation StartExternalRepositoryLoad(
  $namespace: String!
  $repository: String!
  $pluginName: String
  $params: JSON
  $tableParams: [ExternalTableInput!]
  $credentialData: JSON
  $credentialId: String
  $sync: Boolean! = true
  $initialVisibility: RepositoryVisibility! = PUBLIC
) {
  __typename
  startExternalRepositoryLoad(
    namespace: $namespace
    repository: $repository
    pluginName: $pluginName
    params: $params
    tables: $tableParams
    credentialData: $credentialData
    credentialId: $credentialId
    sync: $sync
    initialPermissions: { visibility: $initialVisibility }
  ) {
    taskId
  }
}
"""

GET_PLUGINS = """query ExternalPlugins($seed: String) {
  externalPlugins(seed: $seed) {
    pluginName
    name
    description
    paramsSchema
    credentialsSchema
    tableParamsSchema
    supportsSync
    supportsMount
    supportsLoad
  }
}
"""

GET_PLUGIN = """query ExternalPlugin($pluginName: String!) {
  externalPlugin(name: $pluginName) {
    pluginName
    name
    description
    paramsSchema
    credentialsSchema
    tableParamsSchema
    supportsSync
    supportsMount
    supportsLoad
  }
}"""

START_EXPORT = """mutation StartExport($query: String!) {
  exportQuery(query: $query, exportFormat: "csv") {
    id
  }
}
"""

EXPORT_JOB_STATUS = """query ExportJobStatus($taskId: UUID!) {
  exportJobStatus(taskId: $taskId) {
    taskId
    started
    finished
    status
    userId
    exportFormat
    output
  }
}
"""



import sys
from typing import List, Optional, Tuple

from .keys import ARG_KEYS, ARGUMENT_KEY_MAP


def get_arg_tuples() -> List[Tuple[str, str]]:
    """
    Get the raw (argN, argN+1) tuples from sys.argv

    We could also use click to parse the flags ahead of time, and then pass
    the parsed flag object into SystemConfigGetters. But this way, we avoid
    having to pass down variables from click just to get config values.
    """

    # This one is for when we are on the engine -- then sys.argv doesn't exist in embedded Python
    # and importing the config fails
    sys_args = sys.argv[1:] if hasattr(sys, "argv") else [""]

    return [(a1, a2) for a1, a2 in zip(sys_args, sys_args[1:]) if a1 in ARG_KEYS]


def get_argument_config_value(key: str, default_return: Optional[str] = None) -> Optional[str]:
    """
    Get get the value of an argument, where value is the argument
    immediately following the argument matching a key in ARG_KEYS, e.g.:

        SYS_ARGS = ["--namespace", "foo"]
        --> return "foo"

    Otherwise, return default_return
    """

    arg_tuples = get_arg_tuples()
    matching_values = [v for k, v in arg_tuples if ARGUMENT_KEY_MAP[k] == key]
    num_matching_values = len(matching_values)

    if num_matching_values == 0:
        return default_return

    sys.stderr.write("Warning: multiple values specified for %s \n" % key)
    sys.stderr.write("Using %s \n" % matching_values[0])
    return matching_values[0]
